from builtins import str
from builtins import object
import bcrypt
import logging

from biomaj.mongo_connector import MongoConnector
from biomaj.config import BiomajConfig

class BmajUser(object):
  '''
  Biomaj User
  '''

  def __init__(self, user):

    if MongoConnector.db is None:
      MongoConnector(BiomajConfig.global_config.get('GENERAL','db.url'),
                      BiomajConfig.global_config.get('GENERAL','db.name'))

    self.users = MongoConnector.users
    self.id = user
    self.user = self.users.find_one({'id': user})
    ldap_server = None
    con = None
    if not self.user and BiomajConfig.global_config.get('GENERAL','use_ldap') == '1':
      # Check if in ldap
      #import ldap
      from ldap3 import Server, Connection, AUTH_SIMPLE, STRATEGY_SYNC, STRATEGY_ASYNC_THREADED, SEARCH_SCOPE_WHOLE_SUBTREE, GET_ALL_INFO
      try:
          ldap_host = BiomajConfig.global_config.get('GENERAL','ldap.host')
          ldap_port = BiomajConfig.global_config.get('GENERAL','ldap.port')
          #con = ldap.initialize('ldap://' + ldap_host + ':' + str(ldap_port))
          ldap_server = Server(ldap_host, port = int(ldap_port), get_info = GET_ALL_INFO)
          con = Connection(ldap_server, auto_bind = True, client_strategy = STRATEGY_SYNC, check_names=True)
      except Exception as err:
              logging.error(str(err))
              self.user = None
      ldap_dn = BiomajConfig.global_config.get('GENERAL','ldap.dn')
      base_dn = 'ou=People,' + ldap_dn
      filter = "(&(|(uid=" + user + ")(mail=" + user + ")))"
      try:
          #con.simple_bind_s()
          attrs = ['mail']
          #results = con.search_s(base_dn, ldap.SCOPE_SUBTREE, filter, attrs)
          con.search(base_dn,filter, SEARCH_SCOPE_WHOLE_SUBTREE, attributes=attrs)
          if con.response:
            ldapMail = None
            #for dn, entry in results:
            for r in con.response:
              user_dn = str(r['dn'])
              #if 'mail' not in entry:
              if 'mail' not in r['attributes']:
                logging.error('Mail not set for user '+user)
              else:
                #ldapMail = entry['mail'][0]
                ldapMail = r['attributes']['mail'][0]
            self.user = {
                          'id' : user,
                          'email': ldapMail,
                          'is_ldap': True
                        }
            self.user['_id'] = self.users.insert(self.user)

          else:
            self.user = None
      except Exception as err:
        logging.error(str(err))
      if con:
          con.unbind()

  @staticmethod
  def user_remove(user_name):
    '''
    Remove a user from db

    :param user_name: user name
    :type user_name: str
    '''
    MongoConnector.users.remove({'id': user_name})

  @staticmethod
  def user_banks(user_name):
    '''
    Get user banks name

    :param user_name: user identifier
    :type user_name: str
    :return: list of bank name
    '''
    banks = MongoConnector.banks.find({'properties.owner': user_name}, {'name':1})
    return banks

  @staticmethod
  def list():
    '''
    Get users
    '''
    return MongoConnector.users.find()

  def check_password(self, password):
    if self.user is None:
      return False

    if self.user['is_ldap']:
      #import ldap
      con = None
      ldap_server = None
      #try:
      #    ldap_host = BiomajConfig.global_config.get('GENERAL','ldap.host')
      #    ldap_port = BiomajConfig.global_config.get('GENERAL','ldap.port')
      #    con = ldap.initialize('ldap://' + ldap_host + ':' + str(ldap_port))
      from ldap3 import Server, Connection, AUTH_SIMPLE, STRATEGY_SYNC, STRATEGY_ASYNC_THREADED, SEARCH_SCOPE_WHOLE_SUBTREE, GET_ALL_INFO
      from ldap3.core.exceptions import LDAPBindError
      try:
              ldap_host = BiomajConfig.global_config.get('GENERAL','ldap.host')
              ldap_port = BiomajConfig.global_config.get('GENERAL','ldap.port')
              #con = ldap.initialize('ldap://' + ldap_host + ':' + str(ldap_port))
              ldap_server = Server(ldap_host, port = int(ldap_port), get_info = GET_ALL_INFO)
              con = Connection(ldap_server, auto_bind = True, client_strategy = STRATEGY_SYNC, check_names=True)
      except Exception as err:
              logging.error(str(err))
              return False
      ldap_dn = BiomajConfig.global_config.get('GENERAL','ldap.dn')
      base_dn = 'ou=People,' + ldap_dn
      filter = "(&(|(uid=" + self.user['id'] + ")(mail=" + self.user['id'] + ")))"
      #try:
      #    con.simple_bind_s()
      #except Exception as err:
      #    logging.error(str(err))
      #    return False
      try:
        attrs = ['mail']
        con.search(base_dn,filter, SEARCH_SCOPE_WHOLE_SUBTREE, attributes = attrs)
        #results = con.search_s(base_dn, ldap.SCOPE_SUBTREE, filter, attrs)
        user_dn = None
        ldapMail = None
        ldapHomeDirectory = None
        for r in con.response:
          user_dn = str(r['dn'])
          ldapMail = r['attributes']['mail'][0]
        #for dn, entry in results:
        #    user_dn = str(dn)
        #    ldapMail = entry['mail'][0]
        con.unbind()
        con = Connection(ldap_server, auto_bind = True, read_only=True, client_strategy = STRATEGY_SYNC, user=user_dn, password=password, authentication=AUTH_SIMPLE, check_names=True)
        con.unbind()
        #con.simple_bind_s(user_dn, password)
        #con.unbind_s()
        if user_dn:
          return True
      except LDAPBindError as err:
        logging.error('Bind error: '+str(err))
        return False
      except Exception as err:
        logging.error('Bind error: '+str(err))
        return False

    else:
      hashed = bcrypt.hashpw(password, self.user['hashed_password'])
      if hashed == self.user['hashed_password']:
        return True
      else:
        return False

  def remove(self):
    if self.user is None:
      return False
    self.users.remove({'_id': self.user['_id']})
    return True

  def create(self, password, email=''):
    '''
    Create a new user
    '''
    hashed = bcrypt.hashpw(password, bcrypt.gensalt())
    if self.user is None:
        self.user = {
                      'id' : self.id,
                      'hashed_password': hashed,
                      'email': email,
                      'is_ldap': False
                    }
        self.user['_id'] = self.users.insert(self.user)



from builtins import range
from builtins import object
import threading
import logging
from biomaj.process.metaprocess import MetaProcess

class ProcessFactory(object):
  '''
  Manage process execution
  '''

  NB_THREAD = 2

  def __init__(self, bank):
    self.bank = bank
    self.threads_tasks = []
    if self.bank.session:
      self.meta_data = self.bank.session.get('per_process_metadata')
    else:
      self.meta_data = {}

  def run(self, simulate=False):
    '''
    Run processes

    :param simulate: does not execute process
    :type simulate: bool
    :return: status of execution - bool
    '''
    pass

  def run_threads(self, simulate=False):
    '''
    Start meta threads

    :param simulate: do not execute processes
    :type simulate: bool
    :return: tuple global execution status and status per meta process
    '''
    logging.debug('Start meta threads')
    threads = []
    running_th = []
    for thread_tasks in self.threads_tasks:
      meta_thread = MetaProcess(self.bank, thread_tasks, self.meta_status, self.meta_data, simulate)
      meta_thread._lock = ProcessFactory._LOCK
      meta_thread.workflow = self.workflow
      meta_thread.start()
      threads.append(meta_thread)
      running_th.append(meta_thread)
    # Wait for the end of the threads
    kill_received = False
    while len(running_th) > 0:
      try:
          # Join all threads using a timeout so it doesn't block
          # Filter out threads which have been joined or are None
          running_th = [t.join(1000) for t in running_th if t is not None and t.isAlive()]
      except KeyboardInterrupt:
          logging.warn("Ctrl-c received! Sending kill to threads...")
          logging.warn("Running tasks will continue and process will stop.")
          kill_received = True
          for t in running_th:
              t.kill_received = True

    for meta_thread in threads:
      meta_thread.join()
    global_meta_status = {}
    global_status = True

    for meta_thread in threads:
      for meta in meta_thread.meta_status:
        global_meta_status[meta] = meta_thread.meta_status[meta]
      if not meta_thread.global_status:
        global_status = False

    if kill_received:
      global_status = False

    logging.debug('Meta threads are over')
    return (global_status, global_meta_status)

  def fill_tasks_in_threads(self, metas):
    '''
    Dispatch meta processes in available threads
    '''
    self.threads_tasks = []
    for i in range(0,ProcessFactory.NB_THREAD):
      # Fill array of meta process in future threads
      self.threads_tasks.append([])
    thread_id = 0
    for meta in metas:
      meta_process = meta.strip()
      if thread_id == ProcessFactory.NB_THREAD:
        thread_id = 0
      self.threads_tasks[thread_id].append(meta_process)
      thread_id += 1


class PreProcessFactory(ProcessFactory):
  '''
  Manage preprocesses
  '''

  def __init__(self, bank, metas=None):
    '''
    Creates a preprocess factory

    :param bank: Bank
    :type bank: :class:`biomaj.bank.Bank`
    :param metas: initial status of meta processes
    :type metas: dict
    '''
    ProcessFactory.__init__(self, bank)
    self.meta_status = None
    if metas is not None:
      self.meta_status = metas
    self.workflow = 'preprocess'

  def run(self, simulate=False):
    '''
    Run processes

    :param simulate: does not execute process
    :type simulate: bool
    :return: status of execution - bool
    '''
    logging.info('PROC:PRE')
    if self.bank.config.get('db.pre.process') is None:
      metas = []
    else:
      metas = self.bank.config.get('db.pre.process').split(',')
    self.fill_tasks_in_threads(metas)
    (status, self.meta_status) = self.run_threads(simulate)
    return status

class RemoveProcessFactory(ProcessFactory):
  '''
  Manage remove processes
  '''

  def __init__(self, bank, metas=None):
    '''
    Creates a remove process factory

    :param bank: Bank
    :type bank: :class:`biomaj.bank.Bank`
    :param metas: initial status of meta processes
    :type metas: dict
    '''
    ProcessFactory.__init__(self, bank)
    self.meta_status = None
    if metas is not None:
      self.meta_status = metas
    self.workflow = 'removeprocess'


  def run(self, simulate=False):
    '''
    Run processes

    :param simulate: does not execute process
    :type simulate: bool
    :return: status of execution - bool
    '''
    logging.info('PROC:REMOVE')
    if self.bank.config.get('db.remove.process') is None:
      metas = []
    else:
      metas = self.bank.config.get('db.remove.process').split(',')
    self.fill_tasks_in_threads(metas)
    (status, self.meta_status) = self.run_threads(simulate)
    return status

class PostProcessFactory(ProcessFactory):
  '''
  Manage postprocesses

  self.blocks: dict of meta processes status
  Each meta process status is a dict of process status
  '''

  def __init__(self, bank, blocks=None):
    '''
    Creates a postprocess factory

    :param bank: Bank
    :type bank: :class:`biomaj.bank.Bank`
    :param blocks: initial status of block processes
    :type blocks: dict
    '''
    ProcessFactory.__init__(self, bank)
    self.blocks = {}
    if blocks is not None:
      self.blocks = blocks
    self.workflow = 'postprocess'

  def run(self, simulate=False):
    '''
    Run processes

    :param simulate: does not execute process
    :type simulate: bool
    :return: status of execution - bool
    '''
    logging.info('PROC:POST:BLOCK')
    blocks = self.bank.config.get('BLOCKS')
    if blocks is None or blocks == '':
      process_blocks = []
    else:
      process_blocks = blocks.split(',')
    metas = []
    self.meta_status = None
    global_status = True
    for process_block in process_blocks:
      if not global_status:
        continue
      logging.info('PROC:POST:BLOCK:'+process_block)
      if process_block in self.blocks:
        self.meta_status = self.blocks[process_block]
      # run each block
      metas = self.bank.config.get(process_block.strip()+'.db.post.process').split(',')
      self.fill_tasks_in_threads(metas)
      (status, self.blocks[process_block]) = self.run_threads(simulate)
      if not status:
        global_status = False
    return global_status


ProcessFactory._LOCK = threading.Lock()



from builtins import object
from pymongo import MongoClient


class MongoConnector(object):
  '''
  Connector to mongodb
  '''

  client = None
  db = None
  banks = None
  users = None

  def __init__(self, url, db):
    MongoConnector.client = MongoClient(url)
    MongoConnector.db = MongoConnector.client[db]
    MongoConnector.banks = MongoConnector.db.banks
    MongoConnector.users = MongoConnector.db.users



from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import object
import logging
import logging.config
import os
import configparser
import time
import sys

from biomaj.bmajindex import BmajIndex

class BiomajConfig(object):
  '''
  Manage Biomaj configuration
  '''

  DEFAULTS = {
  'http.parse.dir.line': r'<img[\s]+src="[\S]+"[\s]+alt="\[DIR\]"[\s]*/?>[\s]*<a[\s]+href="([\S]+)/"[\s]*>.*([\d]{2}-[\w\d]{2,5}-[\d]{4}\s[\d]{2}:[\d]{2})',
  'http.parse.file.line': r'<img[\s]+src="[\S]+"[\s]+alt="\[[\s]+\]"[\s]*/?>[\s]<a[\s]+href="([\S]+)".*([\d]{2}-[\w\d]{2,5}-[\d]{4}\s[\d]{2}:[\d]{2})[\s]+([\d\.]+[MKG]{0,1})',
  'http.group.dir.name': 1,
  'http.group.dir.date': 2,
  'http.group.file.name': 1,
  'http.group.file.date': 2,
  'http.group.file.size': 3,
  'visibility.default': 'public',
  'historic.logfile.level': 'INFO',
  'bank.num.threads': 2,
  'files.num.threads': 4,
  'use_elastic': 0,
  'use_drmaa': 0,
  'db.type': '',
  'db.formats': '',
  'keep.old.version': 1,
  'docker.sudo': '1',
  'auto_publish': 0
  }

  # Old biomaj level compatibility
  LOGLEVEL = {
    'DEBUG': logging.DEBUG,
    'VERBOSE': logging.INFO,
    'INFO': logging.INFO,
    'WARN': logging.WARNING,
    'ERR': logging.ERROR
  }

  '''
  Global configuration file
  '''
  global_config = None

  '''
  Per use global configuration file, overriding global_config
  '''
  user_config = None

  @staticmethod
  def load_config(config_file=None, allow_user_config=True):
    '''
    Loads general config

    :param config_file: global.properties file path
    :type config_file: str
    :param allow_user_config: use ~/.biomaj.cfg if present
    :type allow_user_config: bool
    '''
    if config_file is None:
      env_file = os.environ.get('BIOMAJ_CONF')
      if env_file is not None and os.path.exists(env_file):
        config_file = env_file
      else:
        env_file = 'global.properties'
        if os.path.exists(env_file):
          config_file = env_file

    if config_file is None or not os.path.exists(config_file):
      raise Exception('Missing global configuration file')

    BiomajConfig.config_file = os.path.abspath(config_file)

    BiomajConfig.global_config = configparser.SafeConfigParser()

    if allow_user_config and os.path.exists(os.path.expanduser('~/.biomaj.cfg')):
      BiomajConfig.user_config = configparser.SafeConfigParser()
      BiomajConfig.user_config.read([os.path.expanduser('~/.biomaj.cfg')])

    BiomajConfig.global_config.read([config_file])

    # ElasticSearch indexation support
    do_index = False
    if BiomajConfig.global_config.get('GENERAL','use_elastic') and \
      BiomajConfig.global_config.get('GENERAL','use_elastic') == "1":
      do_index = True
    if do_index:
      if BiomajConfig.global_config.get('GENERAL','elastic_nodes'):
        elastic_hosts = BiomajConfig.global_config.get('GENERAL','elastic_nodes').split(',')
      else:
        elastic_hosts = ['localhost']
      elastic_index = BiomajConfig.global_config.get('GENERAL','elastic_index')
      if elastic_index is None:
        elastic_index = 'biomaj'
      BmajIndex.load(index=elastic_index, hosts=elastic_hosts, do_index=do_index)




  def __init__(self, bank, options=None):
    '''
    Loads bank configuration

    :param bank: bank name
    :type bank: str
    :param options: bank options
    :type options: argparse
    '''
    self.name = bank
    if BiomajConfig.global_config is None:
      BiomajConfig.load_config()
    self.config_bank = configparser.SafeConfigParser()
    conf_dir = BiomajConfig.global_config.get('GENERAL', 'conf.dir')
    if not os.path.exists(os.path.join(conf_dir,bank+'.properties')):
      logging.error('Bank configuration file does not exists')
      raise Exception('Configuration file '+bank+'.properties does not exists')
    try:
      self.config_bank.read([os.path.join(conf_dir,bank+'.properties')])
    except Exception as e:
      print("Configuration file error: "+str(e))
      logging.error("Configuration file error "+str(e))
      sys.exit(1)

    self.last_modified = int(os.stat(os.path.join(conf_dir,bank+'.properties')).st_mtime)

    if os.path.exists(os.path.expanduser('~/.biomaj.cfg')):
      logging.config.fileConfig(os.path.expanduser('~/.biomaj.cfg'))
    else:
      logging.config.fileConfig(BiomajConfig.config_file)

    do_log = False
    if options is None:
      do_log = True
    elif hasattr(options,'no_log') and not options.no_log:
      do_log = True
    elif type(options) is dict and 'no_log' in options and not options['no_log']:
      do_log = True

    #if options is None or (( hasattr(options,'no_log') and not options.no_log) or ('no_log' in options and not options['no_log'])):
    if do_log:
      logger = logging.getLogger()
      bank_log_dir = os.path.join(self.get('log.dir'),bank,str(time.time()))
      if not os.path.exists(bank_log_dir):
        os.makedirs(bank_log_dir)
      hdlr = logging.FileHandler(os.path.join(bank_log_dir,bank+'.log'))
      self.log_file = os.path.join(bank_log_dir,bank+'.log')
      if options is not None and options.get_option('log') is not None:
        hdlr.setLevel(BiomajConfig.LOGLEVEL[options.get_option('log')])
      else:
        hdlr.setLevel(BiomajConfig.LOGLEVEL[self.get('historic.logfile.level')])
      formatter = logging.Formatter('%(asctime)s %(levelname)-5.5s [%(name)s][%(threadName)s] %(message)s')
      hdlr.setFormatter(formatter)
      logger.addHandler(hdlr)
    else:
       self.log_file='none'


  def set(self, prop, value, section='GENERAL'):
    self.config_bank.set(section, prop, value)

  def get_bool(self, prop, section='GENERAL', escape=True, default=None):
    '''
    Get a boolean property from bank or general configration. Optionally in section.
    '''
    value = self.get(prop,section,escape,default)
    if value is None:
      return False
    if value is True or value == 'true' or value == '1':
      return True
    else:
      return False

  def get(self, prop, section='GENERAL', escape=True, default=None):
    '''
    Get a property from bank or general configration. Optionally in section.
    '''
    # Compatibility fields
    if prop == 'depends':
      depend = self.get('db.source', section, escape, None)
      if depend:
        return depend

    if self.config_bank.has_option(section,prop):
      val = self.config_bank.get(section,prop)
      if prop == 'remote.dir' and not val.endswith('/'):
        val = val + '/'
      # If regexp, escape backslashes
      if escape and (prop == 'local.files' or prop == 'remote.files' or prop == 'http.parse.dir.line' or prop == 'http.parse.file.line'):
        val = val.replace('\\\\','\\')
      return val

    if BiomajConfig.user_config is not None:
      if BiomajConfig.user_config.has_option(section, prop):
        return BiomajConfig.user_config.get(section, prop)

    if BiomajConfig.global_config.has_option(section, prop):
      return BiomajConfig.global_config.get(section, prop)

    if prop in BiomajConfig.DEFAULTS:
      return BiomajConfig.DEFAULTS[prop]

    return default


  def get_time(self):
    '''
    Return last modification time of config files
    '''
    return self.last_modified



  def check(self):
    '''
    Check configuration
    '''
    status = True
    if not self.get('data.dir'):
      logging.error('data.dir is not set')
      status = False
    if not self.get('conf.dir'):
      logging.error('conf.dir is not set')
      status = False
    if not self.get('log.dir'):
      logging.error('log.dir is not set')
      status = False
    if not self.get('process.dir'):
      logging.error('process.dir is not set')
      status = False
    if not self.get('db.fullname'):
      logging.warn('db.fullname is not set')
    if not self.get('db.formats'):
      logging.warn('db.formats is not set')
    if self.get('use_ldap'):
      if not self.get('ldap.host') or not self.get('ldap.port') or not self.get('ldap.dn'):
        logging.error('use_ldap set to 1 but missing configuration')
        status = False
    if self.get('use_elastic'):
      if not self.get('elastic_nodes') or not self.get('elastic_index'):
        logging.error('use_elastic set to 1 but missing configuration')
        status = False

    if not self.get('celery.queue') or not self.get('celery.broker'):
      logging.warn('celery config is not set, that\'s fine if you do not use Celery for background tasks')

    if not self.get('mail.smtp.host'):
      logging.warn('SMTP mail config not set, you will not be able to send emails')
      status = False
    if self.get('mail.smtp.host') and not self.get('mail.from'):
      logging.error('Mail origin mail.from not set')
      status = False

    if not self.get('offline.dir.name'):
      logging.error('offline.dir.name is not set')
      status = False
    elif self.get('offline.dir.name').startswith('/'):
      logging.error('offline dir must be relative to data.dir and should not start with a /')
      status = False
    if not self.get('dir.version'):
      logging.error('dir.version is not set')
      status = False
    elif self.get('dir.version').startswith('/'):
      logging.error('dir.version must be relative to data.dir and should not start with a /')
      status = False
    if not self.get('protocol'):
      logging.error('protocol is not set')
      status = False
    else:
      protocol = self.get('protocol')
      allowed_protocols = ['multi', 'local', 'ftp', 'sftp', 'http', 'directftp', 'directhttp']
      if protocol not in allowed_protocols:
        logging.error('Protocol not supported: '+protocol)
        status = False
      if protocol != 'multi':
        if protocol != 'local' and not self.get('server'):
          logging.error('server not set')
          status = False
        if not self.get('remote.dir'):
          logging.error('remote.dir not set')
          status = False
        elif not self.get('remote.dir').endswith('/'):
          logging.error('remote.dir must end with a /')
          return False
        if not self.get('remote.files'):
          logging.error('remote.files not set')
          status = False
    if not self.get('local.files'):
      logging.error('local.files is not set')
      status = False
    # Remove processes
    processes = ['db.remove.process','db.pre.process']
    for process in processes:
      if self.get(process):
        metas = self.get(process).split(',')
        for meta in metas:
          if not self.get(meta):
            logging.error('Metaprocess ' + meta + ' not defined')
            status = False
          else:
            procs = self.get(meta).split(',')
            for proc in procs:
              if not self.get(proc+'.name'):
                logging.error('Process '+proc+' not defined')
                status = False
              else:
                if not self.get(proc+'.exe'):
                  logging.error('Process exe for '+proc+' not defined')
                  status = False
    # Check blocks
    if self.get('BLOCKS'):
      blocks = self.get('BLOCKS').split(',')
      for block in blocks:
        if not self.get(block+'.db.post.process'):
          logging.error('Block '+block+' not defined')
          status = False
        else:
          metas = self.get(block+'.db.post.process').split(',')
          for meta in metas:
            if not self.get(meta):
              logging.error('Metaprocess ' + meta + ' not defined')
              status = False
            else:
              procs = self.get(meta).split(',')
              for proc in procs:
                if not self.get(proc+'.name'):
                  logging.error('Process '+proc+' not defined')
                  status = False
                else:
                  if not self.get(proc+'.exe'):
                    logging.error('Process exe for '+proc+' not defined')
                    status = False
    return status



from builtins import str
from builtins import object
import os
import logging
import datetime
import time
import re
import tarfile
import zipfile

from biomaj.utils import Utils

from biomaj.mongo_connector import MongoConnector


class _FakeLock(object):
  '''
  Fake lock for downloaders not called by a Downloadthread
  '''

  def __init__(self):
    pass

  def acquire(self):
    pass

  def release(self):
    pass

class DownloadInterface(object):
  '''
  Main interface that all downloaders must extend
  '''

  files_num_threads = 4

  def __init__(self):
    self.files_to_download = []
    self.files_to_copy = []
    self.error = False
    self.credentials = None
    #bank name
    self.bank = None
    self.mkdir_lock = _FakeLock()
    self.kill_received = False


  def set_progress(self, val, max):
    '''
    Update progress on download

    :param val: number of downloaded files since last progress
    :type val: int
    :param max: number of files to download
    :type max: int
    '''
    logging.debug('Download:progress:'+str(val)+'/'+str(max))
    if not self.bank:
      logging.debug('bank not specified, skipping record of download progress')
      return

    MongoConnector.banks.update({'name': self.bank},{'$inc': {'status.download.progress': val}, '$set': {'status.download.total': max}})

  def match(self, patterns, file_list, dir_list=[], prefix='', submatch=False):
    '''
    Find files matching patterns. Sets instance variable files_to_download.

    :param patterns: regexps to match
    :type patterns: list
    :param file_list: list of files to match
    :type file_list: list
    :param dir_list: sub directories in current dir
    :type dir_list: list
    :param prefix: directory prefix
    :type prefix: str
    :param submatch: first call to match, or called from match
    :type submatch: bool
    '''
    logging.debug('Download:File:RegExp:'+str(patterns))
    if not submatch:
        self.files_to_download = []
    for pattern in patterns:
      subdirs_pattern = pattern.split('/')
      if len(subdirs_pattern) > 1:
        # Pattern contains sub directories
        subdir = subdirs_pattern[0]
        if subdir == '^':
          subdirs_pattern = subdirs_pattern[1:]
          subdir = subdirs_pattern[0]
        if not dir_list and pattern == '**/*':
          # Take all and no more dirs, take all files
          for rfile in file_list:
              rfile['root'] = self.rootdir
              if prefix != '':
                rfile['name'] = prefix + '/' +rfile['name']
              self.files_to_download.append(rfile)
              logging.debug('Download:File:MatchRegExp:'+rfile['name'])
          return
        for direlt in dir_list:
          subdir = direlt['name']
          logging.debug('Download:File:Subdir:Check:'+subdir)
          if pattern == '**/*':
            (subfile_list, subdirs_list) = self.list(prefix+'/'+subdir+'/')
            self.match([pattern], subfile_list, subdirs_list, prefix+'/'+subdir, True)
            for rfile in file_list:
              if pattern == '**/*' or re.match(pattern, rfile['name']):
                rfile['root'] = self.rootdir
                if prefix != '':
                  rfile['name'] = prefix + '/' +rfile['name']
                self.files_to_download.append(rfile)
                logging.debug('Download:File:MatchRegExp:'+rfile['name'])
          else:
            if re.match(subdirs_pattern[0], subdir):
              logging.debug('Download:File:Subdir:Match:'+subdir)
              # subdir match the beginning of the pattern
              # check match in subdir
              (subfile_list, subdirs_list) = self.list(prefix+'/'+subdir+'/')
              self.match(['/'.join(subdirs_pattern[1:])], subfile_list, subdirs_list, prefix+'/'+subdir, True)

      else:
        for rfile in file_list:
          if re.match(pattern, rfile['name']):
            rfile['root'] = self.rootdir
            if prefix != '':
              rfile['name'] = prefix + '/' +rfile['name']
            self.files_to_download.append(rfile)
            logging.debug('Download:File:MatchRegExp:'+rfile['name'])
    if not submatch and len(self.files_to_download) == 0:
      raise Exception('no file found matching expressions')



  def set_permissions(self, file_path, file_info):
    '''
    Sets file attributes to remote ones
    '''
    ftime = datetime.date(int(file_info['year']),int(file_info['month']),int(file_info['day']))
    settime = time.mktime(ftime.timetuple())
    os.utime(file_path, (settime, settime))

  def download_or_copy(self, available_files, root_dir, check_exists=True):
    '''
    If a file to download is available in available_files, copy it instead of downloading it.

    Update the instance variables files_to_download and files_to_copy

    :param available_files: list of files available in root_dir
    :type available files: list
    :param root_dir: directory where files are available
    :type root_dir: str
    :param check_exists: checks if file exists locally
    :type check_exists: bool
    '''

    self.files_to_copy = []
    available_files.sort(key=lambda x: x['name'])
    self.files_to_download.sort(key=lambda x: x['name'])

    new_files_to_download = []

    test1_tuples = ((d['name'], d['year'], d['month'], d['day'], d['size']) for d in self.files_to_download)
    test2_tuples = set((d['name'], d['year'], d['month'], d['day'], d['size']) for d in available_files)
    new_or_modified_files = [t for t in test1_tuples if t not in test2_tuples]
    index = 0

    if len(new_or_modified_files) > 0:
      for file in self.files_to_download:
        if index < len(new_or_modified_files) and \
          file['name'] == new_or_modified_files[index][0]:
          new_files_to_download.append(file)
          index += 1
        else:
          if not check_exists or os.path.exists(os.path.join(root_dir,file['name'])):
            file['root'] = root_dir
            self.files_to_copy.append(file)
          else:
            new_files_to_download.append(file)

    else:
      # Copy everything
      for file in self.files_to_download:
        if not check_exists or os.path.exists(os.path.join(root_dir,file['name'])):
          file['root'] = root_dir
          self.files_to_copy.apppend(file)
        else:
          new_files_to_download.append(file)

    self.files_to_download = new_files_to_download


  def download(self, local_dir):
    '''
    Download remote files to local_dir

    :param local_dir: Directory where files should be downloaded
    :type local_dir: str
    :return: list of downloaded files
    '''
    pass

  def list(self):
    '''
    List directory

    :return: tuple of file list and dir list
    '''
    pass

  def chroot(self, cwd):
    '''
    Change directory
    '''
    pass

  def set_credentials(userpwd):
    '''
    Set credentials in format user:pwd

    :param userpwd: credentials
    :type userpwd: str
    '''
    self.credentials = userpwd

  def close(self):
    '''
    Close connection
    '''
    pass



from future import standard_library
standard_library.install_aliases()
from builtins import str
from builtins import object
import os
import configparser
import logging
import time
import copy

from biomaj.workflow import Workflow, UpdateWorkflow, RemoveWorkflow


class Session(object):
  '''
  BioMAJ bank session
  '''

  OVER = 0

  def __init__(self, name, config, flow=Workflow.FLOW, action='update'):
    '''
    Creates a new session

    :param name: Name of the bank
    :type name: str
    :param config: bank and global config
    :type config: BiomajConfig
    :param flow: Workflow tasks
    :param action: type of flow update|remove
    :type action: str
    :type flow: dict
    '''
    self.name = name
    self.config = config
    self.flow = copy.deepcopy(flow)

    formats = {}
    if self.config.get('db.formats') is not None:
      flist = self.config.get('db.formats').split(',')
      for f in flist:
        formats[f.strip()] = []

    self._session = { 'id':  time.time(),
                      'log_file': self.config.log_file,
                      'status': {},
                      'files': [],
                      'release': None,
                      'remoterelease': None,
                      'formats': formats,
                      'process': {
                                  'postprocess': {},
                                  'preprocess': {},
                                  'removeprocess': {}
                                  },
                      'per_process_metadata': {},
                      'data_dir': self.config.get('data.dir'),
                      'dir_version': self.config.get('dir.version')
                    }
    for flow in self.flow:
        self._session['status'][flow['name']] = False

    self.set('last_modified',self.config.last_modified)

    # Default is update
    self._session['action'] = action

  def reload_postprocess_in_order(self, postprocess):
      '''
      Reloads processes in config order
      '''
      if self.config.get('BLOCKS') is None:
          return postprocess
      copy_postprocess = OrderedDict()
      blocks = self.config.get('BLOCKS').split(',')
      for block in blocks:
          copy_postprocess[block] = OrderedDict()
          metas = self.config.get(block.strip()+'.db.post.process').split(',')
          for meta in metas:
            copy_postprocess[block][meta] = OrderedDict()
            processes = self.config.get(meta.strip()).split(',')
            for process in processes:
                copy_postprocess[block][meta][process] = postprocess[block][meta][process]
      return copy_postprocess

  def reload_in_order(self, otherprocess):
      '''
      Reloads processes in config order
      '''
      if self.config.get(otherprocess.strip()) is None:
          return postprocess
      copy_postprocess = OrderedDict()
      metas = self.config.get(otherprocess.strip()).split(',')
      for meta in metas:
          copy_postprocess[block][meta] = OrderedDict()
          processes = self.config.get(meta.strip()).split(',')
          for process in processes:
              copy_postprocess[block][meta][process] = otherprocess[block][meta][process]
      return copy_postprocess

  def reset_proc(self, type_proc, proc=None):
    '''
    Reset status of processes for type in session

    :param type_proc: postprocess preprocess or removeprocess
    :type type_proc: Workflow.POSTPROCESS, Workflow.PREPROCESS, Workflow.REMOVEPROCESS
    :param proc: reset from block/meta/proc, all reset all
    :type proc: str
    '''
    if type_proc == Workflow.FLOW_POSTPROCESS:
      if proc in self._session['process']['postprocess']:
        self._session['process']['postprocess'] = self.reload_postprocess_in_order(self._session['process']['postprocess'])
        self.reset_meta(self._session['process']['postprocess'][proc])
      else:
        for elt in list(self._session['process']['postprocess'].keys()):
          self.reset_meta(self._session['process']['postprocess'][elt], proc)
    elif type_proc == Workflow.FLOW_PREPROCESS:
      self._session['process']['preprocess'] = self.reload_in_order(self._session['process']['preprocess'])
      self.reset_meta(self._session['process']['preprocess'])
    elif type_proc == Workflow.FLOW_REMOVEPROCESS:
      self._session['process']['removeprocess'] = self.reload_in_order(self._session['process']['removeprocess'])
      self.reset_meta(self._session['process']['removeprocess'], proc)

  def reset_meta(self, metas, proc=None):
    '''
    Reset status of meta processes
    '''
    if proc in metas:
      self.reset_process(procs)
    else:
      for meta in list(metas.keys()):
        self.reset_process(metas[meta], proc)

  def reset_process(self, processes, proc=None):
    '''
    Reset status of processes
    '''
    set_to_false = False
    for process in list(processes.keys()):
      if process == proc or proc is None:
        set_to_false = True
      if set_to_false:
        processes[process] = False


  def load(self, session):
    '''
    Load an existing session
    '''
    self._session = session

  def get_release_directory(self):
    '''
    Get release directroy name
    '''
    return self.name+'-'+str(self._session['release'])

  def get_full_release_directory(self):
    '''
    Get bank directroy for this release
    '''
    #release_dir = os.path.join(self.config.get('data.dir'),
    #              self.config.get('dir.version'),
    #              self.get_release_directory())
    release_dir = os.path.join(self._session['data_dir'],
                  self._session['dir_version'],
                  self.get_release_directory())
    return release_dir

  def get_offline_directory(self):
    '''
    Get bank offline directory
    '''
    return os.path.join(self.config.get('data.dir'),self.config.get('offline.dir.name'))

  def get(self, attr):
    '''
    Return an attribute of session
    '''
    if attr in self._session:
      return self._session[attr]
    else:
      return None

  def set(self, attr, value):
    '''
    Sets an attribute of session
    '''
    self._session[attr] = value

  def get_status(self, status):
    '''
    Return status for a flow event
    '''
    if status not in  self._session['status']:
      return False
    return self._session['status'][status]

  def set_status(self, status, value):
    '''
    Set status for a flow event
    '''
    self._session['status'][status] = value



#!/usr/bin/env python
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import str

import os,sys
#from optparse import OptionParser
import argparse
import pkg_resources
import configparser
import shutil

from biomaj.bank import Bank
from biomaj.config import BiomajConfig
from biomaj.notify import Notify
from biomaj.options import Options
from biomaj.workflow import Workflow

def main():

  parser = argparse.ArgumentParser(add_help=False)
  parser.add_argument('-c', '--config', dest="config",help="Configuration file")
  parser.add_argument('--check', dest="check", help="Check bank property file", action="store_true", default=False)
  parser.add_argument('-u', '--update', dest="update", help="Update action", action="store_true", default=False)
  parser.add_argument('--fromscratch', dest="fromscratch", help="Force a new cycle update", action="store_true", default=False)
  parser.add_argument('-z', '--from-scratch', dest="fromscratch", help="Force a new cycle update", action="store_true", default=False)
  parser.add_argument('-p', '--publish', dest="publish", help="Publish", action="store_true", default=False)
  parser.add_argument('--unpublish', dest="unpublish", help="Unpublish", action="store_true", default=False)

  parser.add_argument('--release', dest="release",help="release of the bank")
  parser.add_argument('--from-task', dest="from_task",help="Start cycle at a specific task (init always executed)")
  parser.add_argument('--process', dest="process",help="Linked to from-task, optionally specify a block, meta or process name to start from")
  parser.add_argument('-l', '--log', dest="log",help="log level")
  parser.add_argument('-r', '--remove', dest="remove", help="Remove a bank release", action="store_true", default=False)
  parser.add_argument('--remove-all', dest="removeall", help="Remove all bank releases and database records", action="store_true", default=False)
  parser.add_argument('-s', '--status', dest="status", help="Get status", action="store_true", default=False)
  parser.add_argument('-b', '--bank', dest="bank",help="bank name")
  parser.add_argument('--owner', dest="owner", help="change owner of the bank")
  parser.add_argument('--stop-before', dest="stop_before",help="Store workflow before task")
  parser.add_argument('--stop-after', dest="stop_after",help="Store workflow after task")
  parser.add_argument('--freeze', dest="freeze", help="Freeze a bank release", action="store_true", default=False)
  parser.add_argument('--unfreeze', dest="unfreeze", help="Unfreeze a bank release", action="store_true", default=False)
  parser.add_argument('-f', '--force', dest="force", help="Force action", action="store_true", default=False)
  parser.add_argument('-h', '--help', dest="help", help="Show usage", action="store_true", default=False)

  parser.add_argument('--search', dest="search", help="Search by format and types", action="store_true", default=False)
  parser.add_argument('--formats', dest="formats",help="List of formats to search, comma separated")
  parser.add_argument('--types', dest="types",help="List of types to search, comma separated")
  parser.add_argument('--query', dest="query",help="Lucene query syntax to search in index")

  parser.add_argument('--show', dest="show", help="Show format files for selected bank", action="store_true", default=False)

  parser.add_argument('-n', '--change-dbname', dest="newbank",help="Change old bank name to this new bank name")
  parser.add_argument('-e', '--move-production-directories', dest="newdir",help="Change bank production directories location to this new path, path must exists")
  parser.add_argument('--visibility', dest="visibility",help="visibility status of the bank")

  parser.add_argument('--maintenance', dest="maintenance",help="Maintenance mode (on/off/status)")

  parser.add_argument('--version', dest="version", help="Show version", action="store_true", default=False)


  options = Options()
  parser.parse_args(namespace=options)

  options.no_log = False

  if options.help:
    print('''
--config: global.properties file path

--status: list of banks with published release
    [OPTIONAL]
    --bank xx / bank: Get status details of bank

--log DEBUG|INFO|WARN|ERR  [OPTIONAL]: set log level in logs for this run, default is set in global.properties file

--check: Check bank property file
    [MANDATORY]
    --bank xx: name of the bank to check (will check xx.properties)

--owner yy: Change owner of the bank (user id)
    [MANDATORY]
    --bank xx: name of the bank

--visibility public|private: change visibility public/private of a bank
    [MANDATORY]
    --bank xx: name of the bank

--change-dbname yy: Change name of the bank to this new name
    [MANDATORY]
    --bank xx: current name of the bank

--move-production-directories yy: Change bank production directories location to this new path, path must exists
    [MANDATORY]
    --bank xx: current name of the bank

--update: Update bank
    [MANDATORY]
    --bank xx: name of the bank(s) to update, comma separated
    [OPTIONAL]
    --publish: after update set as *current* version
    --from-scratch: force a new update cycle, even if release is identical, release will be incremented like (myrel_1)
    --stop-before xx: stop update cycle before the start of step xx
    --stop-after xx: stop update cycle after step xx has completed
    --from-task xx --release yy: Force an re-update cycle for bank release *yy* or from current cycle (in production directories), skipping steps up to *xx*
    --process xx: linked to from-task, optionally specify a block, meta or process name to start from

--publish: Publish bank as current release to use
    [MANDATORY]
    --bank xx: name of the bank to update
    --release xx: release of the bank to publish
--unpublish: Unpublish bank (remove current)
    [MANDATORY]
    --bank xx: name of the bank to update

--remove-all: Remove all bank releases and database records
    [MANDATORY]
    --bank xx: name of the bank to update
    [OPTIONAL]
    --force: remove freezed releases

--remove: Remove bank release (files and database release)
    [MANDATORY]
    --bank xx: name of the bank to update
    --release xx: release of the bank to remove

    Release must not be the *current* version. If this is the case, publish a new release before.

--freeze: Freeze bank release (cannot be removed)
    [MANDATORY]
    --bank xx: name of the bank to update
    --release xx: release of the bank to remove

--unfreeze: Unfreeze bank release (can be removed)
    [MANDATORY]
    --bank xx: name of the bank to update
    --release xx: release of the bank to remove

--search: basic search in bank production releases, return list of banks
   --formats xx,yy : list of comma separated format
  AND/OR
   --types xx,yy : list of comma separated type

   --query "LUCENE query syntax": search in index (if activated)

--show: Show bank files per format
  [MANDATORY]
  --bank xx: name of the bank to show
  [OPTIONAL]
  --release xx: release of the bank to show

--maintenance on/off/status: (un)set biomaj in maintenance mode to prevent updates/removal

    ''')
    return

  if options.version:
    version = pkg_resources.require('biomaj')[0].version
    print('Version: '+str(version))
    return

  bmaj = None
  try:
    if options.config is not None:
      BiomajConfig.load_config(options.config)
    else:
      BiomajConfig.load_config()
  except Exception as e:
    print(str(e))
    sys.exit(1)

  try:

    if options.maintenance:
      if options.maintenance not in ['on', 'off', 'status']:
        print("Wrong maintenance value [on,off,status]")
        sys.exit(1)
      data_dir = BiomajConfig.global_config.get('GENERAL', 'data.dir')
      if BiomajConfig.global_config.has_option('GENERAL', 'lock.dir'):
        lock_dir = BiomajConfig.global_config.get('GENERAL', 'lock.dir')
      else:
        lock_dir = data_dir
      maintenance_lock_file = os.path.join(lock_dir,'biomaj.lock')
      if options.maintenance == 'status':
        if os.path.exists(maintenance_lock_file):
          print("Maintenance: On")
        else:
          print("Maintenance: Off")
        sys.exit(0)
      if options.maintenance == 'on':
        f = open(maintenance_lock_file, 'w')
        f.write('1')
        f.close()
        print("Maintenance set to On")
        sys.exit(0)
      if options.maintenance == 'off':
        if os.path.exists(maintenance_lock_file):
          os.remove(maintenance_lock_file)
        print("Maintenance set to Off")
        sys.exit(0)


    if options.owner:
      if not options.bank:
        print("Bank option is missing")
        sys.exit(1)
      bank = Bank(options.bank, no_log=True)
      bank.set_owner(options.owner)
      sys.exit(0)

    if options.visibility:
      if not options.bank:
        print("Bank option is missing")
        sys.exit(1)
      if options.visibility not in ['public', 'private']:
        print("Valid values are public|private")
        sys.exit(1)
      bank = Bank(options.bank, no_log=True)
      bank.set_visibility(options.visibility)
      print("Do not forget to update accordingly the visibility.default parameter in the configuration file")
      sys.exit(0)

    if options.newdir:
      if not options.bank:
        print("Bank option is missing")
        sys.exit(1)
      if not os.path.exists(options.newdir):
        print("Destination directory does not exists")
      bank = Bank(options.bank, options= options, no_log=True)
      if not bank.bank['production']:
        print("Nothing to move, no production directory")
        sys.exit(0)
      bank.load_session(Workflow.FLOW, None)
      w = Workflow(bank)
      res = w.wf_init()
      if not res:
        sys.exit(1)
      for prod in bank.bank['production']:
        session = bank.get_session_from_release(prod['release'])
        bank.load_session(Workflow.FLOW, session)
        prod_path = bank.session.get_full_release_directory()
        if os.path.exists(prod_path):
          shutil.move(prod_path, options.newdir)
        prod['data_dir'] = options.newdir
      bank.banks.update({'name': options.bank}, {'$set' : { 'production': bank.bank['production'] }})
      print("Bank production directories moved to " + options.newdir)
      print("WARNING: do not forget to update accordingly the data.dir and dir.version properties")
      w.wf_over()
      sys.exit(0)

    if options.newbank:
      if not options.bank:
        print("Bank option is missing")
        sys.exit(1)
      bank = Bank(options.bank, no_log=True)
      conf_dir = BiomajConfig.global_config.get('GENERAL', 'conf.dir')
      bank_prop_file = os.path.join(conf_dir,options.bank+'.properties')
      config_bank = configparser.SafeConfigParser()
      config_bank.read([os.path.join(conf_dir,options.bank+'.properties')])
      config_bank.set('GENERAL', 'db.name', options.newbank)
      newbank_prop_file = open(os.path.join(conf_dir,options.newbank+'.properties'),'w')
      config_bank.write(newbank_prop_file)
      newbank_prop_file.close()
      bank.banks.update({'name': options.bank}, {'$set' : { 'name': options.newbank }})
      os.remove(bank_prop_file)
      print("Bank "+options.bank+" renamed to "+options.newbank)
      sys.exit(0)

    if options.search:
      if options.query:
        res = Bank.searchindex(options.query)
        print("Query matches for :"+options.query)
        print("Release\tFormat\tType\tFiles\n")
        for match in res:
          print(match['_source']['release'] + "\t" + \
                str(match['_source']['format']) + "\t" + \
                str(match['_source']['types']) + "\n")
          for f in match['_source']['files']:
            print("\t\t\t"+f+"\n")
      else:
        formats = []
        if options.formats:
          formats = options.formats.split(',')
        types = []
        if options.types:
          types = options.types.split(',')
        print("Search by formats="+str(formats)+", types="+str(types))
        res = Bank.search(formats, types, False)
        print('#' * 80)
        print("# Name\tRelease")
        for bank in res:
          print(" "+bank['name'])
          for prod in bank['production']:
              iscurrent = ""
              if prod['session'] == bank['current']:
                iscurrent = "current"
              print(" \t"+prod['release']+"\t"+','.join(prod['formats'])+"\t"+','.join(prod['types'])+"\t"+iscurrent)

        print('#' * 80)
        sys.exit(0)

    if options.show:
      if not options.bank:
        print("Bank option is required")
        sys.exit(1)

      bank = Bank(options.bank, no_log=False)
      for prod in bank.bank['production']:
        include = True
        if options.release and (prod['release'] != options.release and prod['prod_dir'] != options.release):
          include =False
        if include:
          session = bank.get_session_from_release(prod['release'])
          print('#' * 80)
          print("# Name:\t"+bank.bank['name'])
          print("# Release:\t"+prod['release'])
          formats = session['formats']
          for fformat in list(formats.keys()):
            print("# \tFormat:\t"+fformat)
            for elt in formats[fformat]:
              print("# \t\tTypes:\t"+','.join(elt['types']))
              print("# \t\tTags:")
              for tag in list(elt['tags'].keys()):
                print("# \t\t\t"+tag+":"+elt['tags'][tag])
              print("# \t\tFiles:")
              for file in elt['files']:
                print("# \t\t\t"+file)
      sys.exit(0)

    if options.check:
      if not options.bank:
        print("Bank name is missing")
        sys.exit(1)
      bank = Bank(options.bank, no_log=False)
      print(options.bank+" check: "+str(bank.check())+"\n")
      sys.exit(0)


    if options.status:
      if options.bank:
        bank = Bank(options.bank)
        _bank = bank.bank
        print('#' * 80)
        print("# Name:\t"+_bank['name'])
        print("# Type:\t"+str(_bank['properties']['type']))
        # Get last update session
        if 'status' in _bank:
          print("# Last update status:\t"+str(_bank['status']['over']['status']))
        release = None
        if 'current' in _bank and _bank['current']:
          for prod in _bank['production']:
            if _bank['current'] == prod['session']:
              release = prod['release']
        print("# Published release:\t"+str(release))
        print("# Production directories")
        for prod in _bank['production']:
          if 'freeze' in prod:
            print("#\tFreeze:\t"+str(prod['freeze']))
          print("#\tRemote release:\t"+prod['remoterelease'])
          print("#\tRelease:\t"+prod['release'])
          print("#\t\tSession:\t"+str(prod['session']))
          release_dir = os.path.join(prod['data_dir'],
                        prod['dir_version'],
                        prod['prod_dir'])
          print("#\t\tDirectory:\t"+release_dir)
        if 'pending' in _bank and len(list(_bank['pending'].keys())) > 0:
          print("# Pending directories")
          for pending in list(_bank['pending'].keys()):
            print("#\tRelease:\t"+pending)
        print('#' * 80)
      else:
        print('#' * 80)
        print("# Name\tType\tRelease")
        banks = Bank.list()
        for bank in banks:
          '''
          production = { 'release': self.session.get('release'),
                          'session': self.session._session['id'],
                          'data_dir': self.config.get('data.dir'),
                          'prod_dir': self.session.get_release_directory()}
          '''
          if 'current' in bank and bank['current']:
            for prod in bank['production']:
              if bank['current'] == prod['session']:
                release = prod['release']
          else:
            release = None
          print(" "+bank['name']+"\t"+','.join(bank['properties']['type'])+"\t"+str(release))
        print('#' * 80)
        sys.exit(0)

    if options.update:
      if not options.bank:
        print("Bank name is missing")
        sys.exit(1)
      banks = options.bank.split(',')
      gres = True
      for bank in banks:
        options.bank = bank
        bmaj = Bank(bank, options)
        print('Log file: '+bmaj.config.log_file)
        res = bmaj.update(depends=True)
        if not res:
          gres = False
        Notify.notifyBankAction(bmaj)
      if not gres:
        sys.exit(1)

    if options.freeze:
      if not options.bank:
        print("Bank name is missing")
        sys.exit(1)
      if not options.release:
        print("Bank release is missing")
        sys.exit(1)
      bmaj = Bank(options.bank, options)
      res = bmaj.freeze(options.release)
      if not res:
        sys.exit(1)

    if options.unfreeze:
      if not options.bank:
        print("Bank name is missing")
        sys.exit(1)
      if not options.release:
        print("Bank release is missing")
        sys.exit(1)
      bmaj = Bank(options.bank, options)
      res = bmaj.unfreeze(options.release)
      if not res:
        sys.exit(1)

    if options.remove or options.removeall:
      if not options.bank:
        print("Bank name is missing")
        sys.exit(1)
      if options.remove and not options.release:
        print("Bank release is missing")
        sys.exit(1)
      if options.removeall:
        bmaj = Bank(options.bank, options, no_log=True)
        print('Log file: '+bmaj.config.log_file)
        res = bmaj.removeAll(options.force)
      else:
        bmaj = Bank(options.bank, options)
        print('Log file: '+bmaj.config.log_file)
        res = bmaj.remove(options.release)
        Notify.notifyBankAction(bmaj)
      if not res:
        sys.exit(1)

    if options.unpublish:
      if not options.bank:
        print("Bank name is missing")
        sys.exit(1)
      bmaj = Bank(options.bank, options)
      bmaj.load_session()
      bmaj.unpublish()
      sys.exit(0)

    if options.publish:
      if not options.bank:
        print("Bank name or release is missing")
        sys.exit(1)
      bmaj = Bank(options.bank, options)
      print('Log file: '+bmaj.config.log_file)
      bmaj.load_session()
      bank = bmaj.bank
      session = None
      if options.get_option('release') is None:
        # Get latest prod release
        if len(bank['production'])>0:
          prod = bank['production'][len(bank['production'])-1]
          for s in bank['sessions']:
            if s['id'] == prod['session']:
              session = s
              break
      else:
        # Search production release matching release
        for prod in bank['production']:
          if prod['release'] == options.release or prod['prod_dir'] == options.release:
            # Search session related to this production release
            for s in bank['sessions']:
              if s['id'] == prod['session']:
                session = s
                break
            break
      if session is None:
        print("No production session could be found for this release")
        sys.exit(1)
      bmaj.session._session = session
      bmaj.publish()
  except Exception as e:
    print(str(e))

if __name__ == '__main__':
    main()



from future import standard_library
standard_library.install_aliases()
from builtins import str
import logging
import pycurl
import io
import re
import os
import datetime

from biomaj.utils import Utils
from biomaj.download.interface import DownloadInterface

class LocalDownload(DownloadInterface):
  '''
  Base class to copy file from local system

  protocol=cp
  server=localhost
  remote.dir=/blast/db/FASTA/

  remote.files=^alu.*\\.gz$

  '''


  def __init__(self, rootdir):
    DownloadInterface.__init__(self)
    logging.debug('Download')
    self.rootdir = rootdir


  def download(self, local_dir):
    '''
    Copy local files to local_dir

    :param local_dir: Directory where files should be copied
    :type local_dir: str
    :return: list of downloaded files
    '''
    logging.debug('Local:Download')
    Utils.copy_files(self.files_to_download, local_dir, lock=self.mkdir_lock)

    return self.files_to_download

  def list(self, directory=''):
    '''
    List FTP directory

    :return: tuple of file and dirs in current directory with details
    '''
    logging.debug('Download:List:'+self.rootdir+directory)
    # lets walk through each line

    rfiles = []
    rdirs = []

    files = [ f for f in os.listdir(self.rootdir + directory) ]
    for file in files:
      rfile = {}
      fstat = os.stat(os.path.join(self.rootdir + directory,file))

      rfile['permissions'] = str(fstat.st_mode)
      rfile['group'] = str(fstat.st_gid)
      rfile['user'] = str(fstat.st_uid)
      rfile['size'] = str(fstat.st_size)
      fstat_mtime = datetime.datetime.fromtimestamp(fstat.st_mtime)
      rfile['month'] = fstat_mtime.month
      rfile['day'] = fstat_mtime.day
      rfile['year'] = fstat_mtime.year
      rfile['name'] = file

      is_dir = False
      if os.path.isdir(os.path.join(self.rootdir + directory,file)):
        is_dir = True

      if not is_dir:
        rfiles.append(rfile)
      else:
        rdirs.append(rfile)
    return (rfiles, rdirs)


  def chroot(self, cwd):
    logging.debug('Download: change dir '+cwd)
    os.chdir(cwd)



from builtins import str
from builtins import range
import os
import logging
import datetime
import time
import re
import threading
import copy
import tarfile
import zipfile
import traceback

class DownloadThread(threading.Thread):

  NB_THREAD = 2


  @staticmethod
  def get_threads(downloader, local_dir):
    '''
    Creates a list of thread for download

    :param downloader: downloader to use
    :type downloader: :class:`biomaj.download.interface.DownloadInterface`
    :param local_dir: directory where files should be downloaded
    :type local_dir: str
    :return: list of threads
    '''
    threads = []
    # Creates threads with copies of the downloader
    for i in range(0,DownloadThread.NB_THREAD):
      new_download = copy.deepcopy(downloader)
      new_download.files_to_download = []
      th = DownloadThread(new_download, local_dir)
      threads.append(th)
    # Now dispatch the files to download to the threads
    thread_id = 0
    for file in downloader.files_to_download:
      if thread_id == DownloadThread.NB_THREAD:
        thread_id = 0
      threads[thread_id].downloader.files_to_download.append(file)
      thread_id += 1
    return threads

  @staticmethod
  def get_threads_multi(downloaders, local_dir):
    '''
    Dispatch multiple downloaders on threads

    :param downloaders: downlaoders to dispatch in threads
    :type downloaders: list of :class:`biomaj.download.interface.DownloadInterface`
    :param local_dir: directory where files should be downloaded
    :type local_dir: str
    :return: list of threads
    '''
    threads = []
    # Creates threads with copies of the downloader
    thread_id = 0
    for downloader in downloaders:
      if thread_id == DownloadThread.NB_THREAD:
        thread_id = 0
      th = DownloadThread(downloader, local_dir)
      threads.append(th)
      thread_id += 1
    return threads

  def __init__(self, downloader, local_dir):
    '''
    Download thread to download a list of files

    :param downloader: downloader to use
    :type downloader: :class:`biomaj.download.interface.DownloadInterface`
    :param local_dir: directory to download files
    :type local_dir: str
    '''
    threading.Thread.__init__(self)
    self.downloader = downloader
    self.downloader.mkdir_lock = DownloadThread.MKDIR_LOCK
    self.downloader.kill_received = False
    self.local_dir = local_dir
    self.error = False
    self._stopevent = threading.Event( )

  def run(self):
    logging.info('Start download thread')
    if self.downloader is None:
      return True
    self.error = False
    try:
      self.downloader.download(self.local_dir)
      self.downloader.close()
    except Exception as e:
      logging.error('Error in download execution of thread: '+str(e))
      logging.debug(traceback.format_exc())
      self.error = True

  def stop(self):
    self._stopevent.set( )


DownloadThread.MKDIR_LOCK = threading.Lock()



from builtins import str
from builtins import object
import os
import logging
import time
import shutil
import copy

from biomaj.mongo_connector import MongoConnector

from biomaj.session import Session
from biomaj.workflow import UpdateWorkflow, RemoveWorkflow, Workflow
from biomaj.config import BiomajConfig
from biomaj.options import Options
from biomaj.process.processfactory import ProcessFactory
from biomaj.bmajindex import BmajIndex

#from bson.objectid import ObjectId


class Bank(object):
  '''
  BioMAJ bank
  '''

  def __init__(self, name, options=None, no_log=False):
    '''
    Get a bank from db or creates a new one

    :param name: name of the bank, must match its config file
    :type name: str
    :param options: bank options
    :type options: argparse
    :param no_log: create a log file for the bank
    :type no_log: bool
    '''
    logging.debug('Initialize '+name)
    if BiomajConfig.global_config is None:
      raise Exception('Configuration must be loaded first')

    self.name = name
    self.depends = []
    self.no_log = no_log

    if no_log:
      if options is None:
        #options = {'no_log': True}
        options = Options()
        options.no_log = True
      else:
        options.no_log = no_log

    self.config = BiomajConfig(self.name, options)

    if self.config.get('bank.num.threads') is not None:
      ProcessFactory.NB_THREAD = int(self.config.get('bank.num.threads'))

    if self.config.log_file is not None and self.config.log_file != 'none':
      logging.info("Log file: "+self.config.log_file)

    #self.options = Options(options)
    if options is None:
      self.options = Options()
    else:
      self.options = options

    if MongoConnector.db is None:
      MongoConnector(BiomajConfig.global_config.get('GENERAL','db.url'),
                      BiomajConfig.global_config.get('GENERAL','db.name'))

    self.banks = MongoConnector.banks

    self.bank = self.banks.find_one({'name': self.name})

    if self.bank is None:
        self.bank = {
                      'name' : self.name,
                      'current': None,
                      'sessions': [],
                      'production': [],
                      'properties': self.get_properties()
                    }
        self.bank['_id'] = self.banks.insert(self.bank)

    self.session = None
    self.use_last_session = False

  def check(self):
    '''
    Checks bank configuration
    '''
    return self.config.check()

    '''
    offline.dir.name=offline/ncbi/blast/alu_tmp
    dir.version=ncbi/blast/alu
    protocol=ftp
    server=ftp.ncbi.nih.gov
    remote.dir=/blast/db/FASTA/

    remote.files=^alu.*\.gz$
    local.files=^alu\.(a|n).*
    '''

    return 'Not yet implemented'

  def is_locked(self):
    '''
    Checks if bank is locked ie action is in progress
    '''
    data_dir = self.config.get('data.dir')
    lock_dir = self.config.get('lock.dir',default=data_dir)
    lock_file = os.path.join(lock_dir,self.name+'.lock')
    if os.path.exists(lock_file):
      return True
    else:
      return False

  def get_bank(self):
    '''
    Get bank stored in db

    :return: bank json object
    '''
    return self.bank

  @staticmethod
  def get_banks_disk_usage():
    '''
    Get disk usage per bank and release
    '''
    if MongoConnector.db is None:
      MongoConnector(BiomajConfig.global_config.get('GENERAL','db.url'),
                      BiomajConfig.global_config.get('GENERAL','db.name'))

    bank_list = []
    banks = MongoConnector.banks.find({},{'name': 1, 'production': 1})
    for b in banks:
      bank_elt = {'name': b['name'], 'size': 0, 'releases': []}
      for p in b['production']:
        if p['size'] is None:
          p['size'] = 0
        bank_elt['size'] += p['size']
        bank_elt['releases'].append({'name': p['release'], 'size': p['size']})
      bank_list.append(bank_elt)
    return bank_list

  def update_dependencies(self):
    '''
    Update bank dependencies

    :return: status of updates
    '''
    self.depends = []
    if self.run_depends:
      depends = self.get_dependencies()
    else:
      depends = []

    self.session.set('depends', {})
    res = True
    for dep in depends:
      self.session._session['depends'][dep] = False
    for dep in depends:
      if self.session._session['depends'][dep]:
        logging.debug('Update:Depends:'+dep+':SKIP')
        # Bank has been marked as depends multiple times, run only once
        continue
      logging.info('Update:Depends:'+dep)
      b = Bank(dep)
      res = b.update()
      self.depends.append(b)
      self.session._session['depends'][dep] = res
      logging.info('Update:Depends:'+dep+':'+str(res))
      if not res:
        break
    return res

  def get_bank(self, bank, no_log=False):
    '''
    Gets an other bank
    '''
    return Bank(bank, no_log=no_log)

  def get_dependencies(self, bank=None):
    '''
    Search all bank dependencies

    :return: list of bank names to update
    '''
    if bank is None:
      deps = self.config.get('depends')
    else:
      deps = bank.config.get('depends')
    if deps is None:
      return []
    # Mainn deps
    deps = deps.split(',')
    # Now search in deps if they themselves depend on other banks
    for dep in deps:
      b = Bank(dep)
      deps = b.get_dependencies() + deps
    return deps


  def is_owner(self):
    '''
    Checks if current user is owner or admin
    '''
    admin_config = self.config.get('admin')
    admin = []
    if admin_config is not None:
      admin = [x.strip() for x in admin_config.split(',')]
    if admin and os.environ['LOGNAME'] in admin:
      return True
    if os.environ['LOGNAME'] == self.bank['properties']['owner']:
      return True
    return False

  def set_owner(self, owner):
    '''
    Update bank owner, only if current owner
    '''
    if not self.is_owner():
      logging.error('Not authorized, bank owned by '+self.bank['properties']['owner'])
      raise Exception('Not authorized, bank owned by '+self.bank['properties']['owner'])

    self.banks.update({'name': self.name}, {'$set' : { 'properties': { 'owner': owner} }})

  def set_visibility(self, visibility):
    '''
    Update bank visibility, only if current owner
    '''
    if not self.is_owner():
      logging.error('Not authorized, bank owned by '+self.bank['properties']['owner'])
      raise Exception('Not authorized, bank owned by '+self.bank['properties']['owner'])

    self.banks.update({'name': self.name}, {'$set' : { 'properties': { 'visibility': visibility} }})



  def get_properties(self):
    '''
    Read bank properties from config file

    :return: properties dict
    '''

    owner = os.environ['LOGNAME']
    # If owner not set, use current user
    if self.bank:
        owner = self.bank['properties']['owner']

    props = {
      'visibility': self.config.get('visibility.default'),
      'type': self.config.get('db.type').split(','),
      'tags': [],
      'owner': owner
    }

    return props

  @staticmethod
  def searchindex(query):
    return BmajIndex.searchq(query)


  @staticmethod
  def search(formats=[], types=[], with_sessions=True):
    '''
    Search all bank releases matching some formats and types

    Matches production release with at least one of formats and one of types
    '''
    if MongoConnector.db is None:
      MongoConnector(BiomajConfig.global_config.get('GENERAL','db.url'),
                      BiomajConfig.global_config.get('GENERAL','db.name'))
    filter = {}
    if formats:
      filter['production.formats'] = {'$in': formats}
    if with_sessions:
      res = MongoConnector.banks.find(filter)
    else:
      res = MongoConnector.banks.find(filter,{'sessions': 0})
    # Now search in which production release formats and types apply
    search_list = []
    for r in res:
      prod_to_delete = []
      for p in r['production']:
        is_format = False
        if not formats:
          is_format = True
        # Are formats present in this production release?
        for f in formats:
          if f in p['formats']:
            is_format = True
            break
        # Are types present in this production release?
        is_type = False
        if not types:
          is_type = True
        if is_format:
          for t in types:
            if t in p['types'] or t in r['properties']['type']:
              is_type = True
              break
        if not is_type or not is_format:
          prod_to_delete.append(p)
      for prod_del in prod_to_delete:
        r['production'].remove(prod_del)
      if len(r['production'])>0:
        search_list.append(r)
    return search_list

  @staticmethod
  def list(with_sessions=False):
    '''
    Return a list of banks

    :param with_sessions: should sessions be returned or not (can be quite big)
    :type with_sessions: bool
    :return: list of :class:`biomaj.bank.Bank`
    '''
    if MongoConnector.db is None:
      MongoConnector(BiomajConfig.global_config.get('GENERAL','db.url'),
                      BiomajConfig.global_config.get('GENERAL','db.name'))


    bank_list = []
    if with_sessions:
      res = MongoConnector.banks.find({})
    else:
      res = MongoConnector.banks.find({},{'sessions': 0})
    for r in res:
      bank_list.append(r)
    return bank_list

  def controls(self):
    '''
    Initial controls (create directories etc...)
    '''
    data_dir = self.config.get('data.dir')
    bank_dir = self.config.get('dir.version')
    bank_dir = os.path.join(data_dir,bank_dir)
    if not os.path.exists(bank_dir):
      os.makedirs(bank_dir)

    offline_dir = self.config.get('offline.dir.name')
    offline_dir = os.path.join(data_dir,offline_dir)
    if not os.path.exists(offline_dir):
      os.makedirs(offline_dir)

    log_dir = self.config.get('log.dir')
    log_dir = os.path.join(log_dir,self.name)
    if not os.path.exists(log_dir):
      os.makedirs(log_dir)


  def _delete(self):
    '''
    Delete bank from database, not files
    '''
    self.banks.remove({'_id': self.bank['_id']})

  def save_session(self):
    '''
    Save session in database
    '''
    self.session._session['last_update_time'] = time.time()
    self.session._session['log_file'] = self.config.log_file
    if self.use_last_session:
      # Remove last session
      self.banks.update({'name': self.name}, {'$pull' : { 'sessions': { 'id': self.session._session['id']} }})
    # Insert session
    if self.session.get('action') == 'update':
      action = 'last_update_session'
    if self.session.get('action') == 'remove':
      action = 'last_remove_session'
    self.banks.update({'name': self.name}, {
      '$set': {
        action: self.session._session['id'],
        'properties': self.get_properties()
      },
      '$push' : { 'sessions': self.session._session }
      })
    BmajIndex.add(self.name, self.session._session)
    if self.session.get('action') == 'update' and not self.session.get_status(Workflow.FLOW_OVER) and self.session.get('release'):
      self.banks.update({'name': self.name},{'$set': {'pending.'+self.session.get('release'): self.session._session['id']}})
    if self.session.get('action') == 'update' and self.session.get_status(Workflow.FLOW_OVER) and self.session.get('update'):
      # We expect that a production release has reached the FLOW_OVER status.
      # If no update is needed (same release etc...), the *update* session of the session is set to False
      logging.debug('Bank:Save:'+self.name)
      if len(self.bank['production']) > 0:
        # Remove from database
        self.banks.update({'name': self.name}, {'$pull' : { 'production': { 'release': self.session._session['release'] }}})
        # Update local object
        #index = 0
        #for prod in self.bank['production']:
        #  if prod['release'] == self.session._session['release']:
        #    break;
        #  index += 1
        #if index < len(self.bank['production']):
        #  self.bank['production'].pop(index)
      release_types = []
      for release_format in self.session._session['formats']:
        for release_files in self.session._session['formats'][release_format]:
          if release_files['types']:
            for rtype in release_files['types']:
              if rtype not in release_types:
                release_types.append(rtype)
      prod_dir = self.session.get_release_directory()
      if self.session.get('prod_dir'):
        prod_dir = self.session.get('prod_dir')
      production = { 'release': self.session.get('release'),
                      'remoterelease': self.session.get('remoterelease'),
                      'session': self.session._session['id'],
                      'formats': list(self.session._session['formats'].keys()),
                      'types': release_types,
                      'size': self.session.get('fullsize'),
                      'data_dir': self.session._session['data_dir'],
                      'dir_version': self.session._session['dir_version'],
                      'prod_dir': prod_dir,
                      'freeze': False }
      self.bank['production'].append(production)

      self.banks.update({'name': self.name},
                        {'$push': {'production': production},
                         '$unset': {'pending.'+self.session.get('release'): ''}
                        })

      #self.banks.update({'name': self.name},
      #                  {'$unset': 'pending.'+self.session.get('release')
      #                  })

    self.bank = self.banks.find_one({'name': self.name})

  def clean_old_sessions(self):
    '''
    Delete old sessions, not latest ones nor related to production sessions
    '''
    if self.session is None:
      return
    # No previous session
    if 'sessions' not in self.bank:
      return
    #'last_update_session' in self.bank and self.bank['last_update_session']
    old_sessions = []
    prod_releases = []
    for session in self.bank['sessions']:
      if session['id'] == self.session.get('id'):
        # Current session
        prod_releases.append(session['release'])
        continue
      if session['id'] == self.session.get('last_update_session'):
        prod_releases.append(session['release'])
        continue
      if session['id'] == self.session.get('last_remove_session'):
        continue
      is_prod_session = False
      for prod in self.bank['production']:
        if session['id'] == prod['session']:
          is_prod_session = True
          break
      if is_prod_session:
        prod_releases.append(session['release'])
        continue
      old_sessions.append(session)
    if len(old_sessions) > 0:
      for session in old_sessions:
        session_id = session['id']
        self.banks.update({'name': self.name}, {'$pull' : { 'sessions': { 'id': session_id }}})
        if session['release'] not in prod_releases and session['release'] != self.session.get('release'):
          # There might be unfinished releases linked to session, delete them
          # if they are not related to a production directory or latest run
          session_dir = os.path.join(self.config.get('data.dir'),
                                      self.config.get('dir.version'),
                                      self.name+'-'+str(session['release']))
          if os.path.exists(session_dir):
            logging.info('Bank:DeleteOldSessionDir:'+self.name+'-'+str(session['release']))
            shutil.rmtree(session_dir)
      self.bank = self.banks.find_one({'name': self.name})

  def publish(self):
    '''
    Set session release to *current*
    '''
    if not self.is_owner():
      logging.error('Not authorized, bank owned by '+self.bank['properties']['owner'])
      raise Exception('Not authorized, bank owned by '+self.bank['properties']['owner'])

    current_link = os.path.join(self.config.get('data.dir'),
                                self.config.get('dir.version'),
                                'current')
    prod_dir = self.session.get_full_release_directory()

    to_dir = os.path.join(self.config.get('data.dir'),
                  self.config.get('dir.version'))

    if os.path.lexists(current_link):
      os.remove(current_link)
    os.chdir(to_dir)
    os.symlink(self.session.get_release_directory(),'current')
    self.bank['current'] = self.session._session['id']
    self.banks.update({'name': self.name},
                      {
                      '$set': {'current': self.session._session['id']}
                      })

  def unpublish(self):
    '''
    Unset *current*
    '''
    if not self.is_owner():
      logging.error('Not authorized, bank owned by '+self.bank['properties']['owner'])
      raise Exception('Not authorized, bank owned by '+self.bank['properties']['owner'])

    current_link = os.path.join(self.config.get('data.dir'),
                                self.config.get('dir.version'),
                                'current')

    if os.path.lexists(current_link):
      os.remove(current_link)
    self.banks.update({'name': self.name},
                      {
                      '$set': {'current': None}
                      })


  def get_production(self, release):
    '''
    Get production field for release

    :param release: release name or production dir name
    :type release: str
    :return: production field
    '''
    production = None
    for prod in self.bank['production']:
      if prod['release'] == release or prod['prod_dir'] == release:
        production = prod
    return production

  def freeze(self, release):
    '''
    Freeze a production release

    When freezed, a production release cannot be removed (manually or automatically)

    :param release: release name or production dir name
    :type release: str
    :return: bool
    '''
    if not self.is_owner():
      logging.error('Not authorized, bank owned by '+self.bank['properties']['owner'])
      raise Exception('Not authorized, bank owned by '+self.bank['properties']['owner'])


    rel = None
    for prod in self.bank['production']:
      if prod['release'] == release or prod['prod_dir'] == release:
        # Search session related to this production release
        rel = prod['release']
    if rel is None:
      logging.error('Release not found: '+release)
    self.banks.update({'name': self.name, 'production.release': rel},{'$set': { 'production.$.freeze': True }})
    self.bank = self.banks.find_one({'name': self.name})
    return True

  def unfreeze(self, release):
    '''
    Unfreeze a production release to allow removal

    :param release: release name or production dir name
    :type release: str
    :return: bool
    '''
    if not self.is_owner():
      logging.error('Not authorized, bank owned by '+self.bank['properties']['owner'])
      raise Exception('Not authorized, bank owned by '+self.bank['properties']['owner'])

    rel = None
    for prod in self.bank['production']:
      if prod['release'] == release or prod['prod_dir'] == release:
        # Search session related to this production release
        rel = prod['release']
    if rel is None:
      logging.error('Release not found: '+release)
    self.banks.update({'name': self.name, 'production.release': rel},{'$set': { 'production.$.freeze': False }})
    self.bank = self.banks.find_one({'name': self.name})
    return True

  def get_new_session(self, flow=Workflow.FLOW):
    '''
    Returns an empty session

    :param flow: kind of workflow
    :type flow: :func:`biomaj.workflow.Workflow.FLOW`
    '''
    return Session(self.name, self.config, flow)

  def get_session_from_release(self, release):
    '''
    Loads the session matching a specific release

    :param release: release name oe production dir
    :type release: str
    :return: :class:`biomaj.session.Session`
    '''
    oldsession = None
    # Search production release matching release
    for prod in self.bank['production']:
      if prod['release'] == release or prod['prod_dir'] == release:
        # Search session related to this production release
        for s in self.bank['sessions']:
          if s['id'] == prod['session']:
            oldsession = s
            break
        break
    if oldsession is None:
      # No prod session, try to find a session for this release, session may have failed or be stopped
      for s in self.bank['sessions']:
        if s['release'] and release.endswith(s['release']):
          oldsession = s
    if oldsession is None:
      logging.error('No production session could be found for this release')
    return oldsession

  def load_session(self, flow=Workflow.FLOW, session=None):
    '''
    Loads last session or, if over or forced, a new session

    Creates a new session or load last session if not over

    :param flow: kind of workflow
    :type flow: :func:`biomaj.workflow.Workflow.FLOW`
    '''
    if session is not None:
      logging.debug('Load specified session '+str(session['id']))
      self.session = Session(self.name, self.config, flow)
      self.session.load(session)
      self.use_last_session = True
      return
    if len(self.bank['sessions']) == 0 or self.options.get_option(Options.FROMSCRATCH):
        self.session = Session(self.name, self.config, flow)
        logging.debug('Start new session')
    else:
        # Take last session
        self.session = Session(self.name, self.config, flow)
        session_id = None
        # Load previous session for updates only
        if self.session.get('action') == 'update' and 'last_update_session' in self.bank and self.bank['last_update_session']:
          session_id = self.bank['last_update_session']
          load_session = None
          for session in self.bank['sessions']:
            if session['id'] == session_id:
              load_session = session
              break
          if load_session is not None:
            #self.session.load(self.bank['sessions'][len(self.bank['sessions'])-1])
            self.session.load(session)
            #if self.config.last_modified > self.session.get('last_modified'):
            #  # Config has changed, need to restart
            #  self.session = Session(self.name, self.config, flow)
            #  logging.info('Configuration file has been modified since last session, restart in any case a new session')
            if self.session.get_status(Workflow.FLOW_OVER) and self.options.get_option(Options.FROM_TASK) is None:
              previous_release = self.session.get('remoterelease')
              self.session = Session(self.name, self.config, flow)
              self.session.set('previous_release', previous_release)
              logging.debug('Start new session')
            else:
              logging.debug('Load previous session '+str(self.session.get('id')))
              self.use_last_session = True


  def remove_session(self, sid):
    '''
    Delete a session from db

    :param sid: id of the session
    :type sid: long
    :return: bool
    '''
    session_release = None
    _tmpbank = self.banks.find_one({'name': self.name})
    for s in _tmpbank['sessions']:
      if s['id'] == sid:
        session_release = s['release']
    if session_release is not None:
      self.banks.update({'name': self.name},{'$pull':{
                                            'sessions': {'id': sid},
                                            'production': {'session': sid}
                                            },
                                            '$unset': {
                                              'pending.'+session_release: ''
                                            }
                      })
    else:
      self.banks.update({'name': self.name},{'$pull':{
                                            'sessions': {'id': sid},
                                            'production': {'session': sid}
                                            }
                      })
    # Update object
    self.bank = self.banks.find_one({'name': self.name})
    if session_release is not None:
      BmajIndex.remove(self.name, session_release)
    return True

  def get_data_dir(self):
    '''
    Returns bank data directory

    :return: str
    '''
    return os.path.join(self.config.get('data.dir'),
                      self.config.get('dir.version'))

  def removeAll(self, force=False):
    '''
    Remove all bank releases and database records

    :param force: force removal even if some production dirs are freezed
    :type force: bool
    :return: bool
    '''
    if not force:
      has_freeze = False
      for prod in self.bank['production']:
        if 'freeze' in prod and prod['freeze']:
          has_freeze = True
          break
      if has_freeze:
        logging.error('Cannot remove bank, some production directories are freezed, use force if needed')
        return False

    self.banks.remove({'name': self.name})
    BmajIndex.delete_all_bank(self.name)
    bank_data_dir = self.get_data_dir()
    logging.warn('DELETE '+bank_data_dir)
    if os.path.exists(bank_data_dir):
      shutil.rmtree(bank_data_dir)
    bank_offline_dir = os.path.join(self.config.get('data.dir'),self.config.get('offline.dir.name'))
    if os.path.exists(bank_offline_dir):
      shutil.rmtree(bank_offline_dir)
    bank_log_dir = os.path.join(self.config.get('log.dir'),self.name)
    if os.path.exists(bank_log_dir) and self.no_log:
      shutil.rmtree(bank_log_dir)
    return True

  def get_status(self):
    '''
    Get status of current workflow

    :return: dict of current workflow status
    '''
    if self.bank['status'] is None:
      return {}
    return self.bank['status']

  def remove(self, release):
    '''
    Remove a release (db and files)

    :param release: release or release directory
    :type release: str
    :return: bool
    '''
    logging.warning('Bank:'+self.name+':Remove')

    if not self.is_owner():
      logging.error('Not authorized, bank owned by '+self.bank['properties']['owner'])
      raise Exception('Not authorized, bank owned by '+self.bank['properties']['owner'])

    self.session = self.get_new_session(RemoveWorkflow.FLOW)
    oldsession = None
    # Search production release matching release
    for prod in self.bank['production']:
      if prod['release'] == release or prod['prod_dir'] == release:
        if 'freeze' in prod and prod['freeze']:
          logging.error('Cannot remove release, release is freezed, unfreeze it first')
          return False
        # Search session related to this production release
        for s in self.bank['sessions']:
          if s['id'] == prod['session']:
            oldsession = s
            break
        break
    if oldsession is None:
      logging.error('No production session could be found for this release')
      return False
    if 'current' in self.bank and self.bank['current'] == oldsession['id']:
      logging.error('This release is the release in the main release production, you should first unpublish it')
      return False

    # New empty session for removal
    session = Session(self.name, self.config, RemoveWorkflow.FLOW)
    session.set('action', 'remove')
    session.set('release', oldsession['release'])
    session.set('update_session_id', oldsession['id'])
    self.session = session
    # Reset status, we take an update session
    res = self.start_remove(session)
    self.save_session()

    return res

  def update(self, depends=False):
    '''
    Launch a bank update

    :param depends: run update of bank dependencies first
    :type depends: bool
    :return: bool
    '''
    logging.warning('Bank:'+self.name+':Update')

    if not self.is_owner():
      logging.error('Not authorized, bank owned by '+self.bank['properties']['owner'])
      raise Exception('Not authorized, bank owned by '+self.bank['properties']['owner'])

    self.run_depends = depends

    self.controls()
    if self.options.get_option('release'):
      logging.info('Bank:'+self.name+':Release:'+self.options.get_option('release'))
      s = self.get_session_from_release(self.options.get_option('release'))
      # No session in prod
      if s is None:
        logging.error('Release does not exists: '+self.options.get_option('release'))
        return False
      self.load_session(UpdateWorkflow.FLOW, s)
    else:
      logging.info('Bank:'+self.name+':Release:latest')
      self.load_session(UpdateWorkflow.FLOW)
    # if from task, reset workflow status in session.
    if self.options.get_option('from_task'):
      set_to_false = False
      for task in self.session.flow:
        # If task was in False status (KO) and we ask to start after this task, exit
        if not set_to_false and not self.session.get_status(task['name']) and  task['name'] != self.options.get_option('from_task'):
          logging.error('Previous task '+task['name']+' was not successful, cannot restart after this task')
          return False
        if task['name'] == self.options.get_option('from_task'):
          set_to_false = True
        if set_to_false:
          # After from_task task, tasks must be set to False to be run
          self.session.set_status(task['name'], False)
          proc = None
          if task['name'] in [Workflow.FLOW_POSTPROCESS, Workflow.FLOW_PREPROCESS, Workflow.FLOW_REMOVEPROCESS]:
            proc = self.options.get_option('process')
            self.session.reset_proc(task['name'], proc)
          #if task['name'] == Workflow.FLOW_POSTPROCESS:
          #  self.session.reset_proc(Workflow.FLOW_POSTPROCESS, proc)
          #elif task['name'] == Workflow.FLOW_PREPROCESS:
          #  self.session.reset_proc(Workflow.FLOW_PREPROCESS, proc)
          #elif task['name'] == Workflow.FLOW_REMOVEPROCESS:
          #  self.session.reset_proc(Workflow.FLOW_REMOVEPROCESS, proc)
    self.session.set('action','update')
    res = self.start_update()
    self.save_session()
    return res

  def start_remove(self, session):
    '''
    Start a removal workflow

    :param session: Session to remove
    :type session: :class:`biomaj.session.Session`
    :return: bool
    '''
    workflow = RemoveWorkflow(self, session)
    return workflow.start()

  def start_update(self):
    '''
    Start an update workflow
    '''
    workflow = UpdateWorkflow(self)
    return workflow.start()



from builtins import str
from builtins import object
import logging
import os
import subprocess
import tempfile

class Process(object):
  '''
  Define a process to execute
  '''

  def __init__(self, name, exe, args, desc=None, proc_type=None, expand=True, bank_env=None, log_dir=None):
    '''
    Define one process

    :param name: name of the process (descriptive)
    :type name: str
    :param exe: path to the executable (relative to process.dir or full path)
    :type exe: str
    :param args: arguments
    :type args: str
    :param desc: process description
    :type desc: str
    :param proc_type: types of data generated by process
    :type proc_type: str
    :param expand: allow shell expansion on command line
    :type expand: bool
    :param bank_env: environnement variables to set
    :type bank_env: list
    :param log_dir: directroy to place process stdout and stderr
    :type log_dir: str
    '''
    # Replace env vars in args
    if args:
      for key,value in bank_env.items():
        if value is not None:
          args = args.replace('${'+key+'}', value)

    self.name = name
    self.exe = exe
    self.desc= desc
    if args is not None:
      self.args = args.split()
    else:
      self.args = []
    self.bank_env = bank_env
    self.type = proc_type
    self.expand = expand
    if log_dir is not None:
      self.output_file = os.path.join(log_dir,name+'.out')
      self.error_file = os.path.join(log_dir,name+'.err')
    else:
      self.output_file = name+'.out'
      self.error_file = name+'.err'

    self.types = ''
    self.format = ''
    self.tags = ''
    self.files = ''

  def run(self, simulate=False):
    '''
    Execute process

    :param simulate: does not execute process
    :type simulate: bool
    :return: exit code of process
    '''
    args = [ self.exe ] + self.args
    logging.debug('PROCESS:EXEC:'+str(self.args))
    err= False
    if not simulate:
      logging.info('PROCESS:RUN:'+self.name)
      with open(self.output_file,'w') as fout:
        with open(self.error_file,'w') as ferr:
          if self.expand:
            args = " ".join(args)
            proc = subprocess.Popen(args, stdout=fout, stderr=ferr, env=self.bank_env, shell=True)
          else:
            proc = subprocess.Popen(args, stdout=fout, stderr=ferr, env=self.bank_env, shell=False)
          proc.wait()
          if proc.returncode == 0:
            err = True
          else:
            logging.error('PROCESS:ERROR:'+self.name)
          fout.flush()
          ferr.flush()
    else:
      err = True
    logging.info('PROCESS:EXEC:' + self.name + ':' + str(err))

    return err

class DockerProcess(Process):
  def __init__(self, name, exe, args, desc=None, proc_type=None, docker=None, expand=True, bank_env=None, log_dir=None, use_sudo=True):
    Process.__init__(self, name, exe, args, desc, proc_type, expand, bank_env, log_dir)
    self.docker = docker
    self.use_sudo = use_sudo

  def run(self, simulate=False):
    '''
    Execute process in docker container

    :param simulate: does not execute process
    :type simulate: bool
    :return: exit code of process
    '''
    use_sudo = ''
    if self.use_sudo:
      use_sudo = 'sudo'
    release_dir = self.bank_env['datadir']+'/'+self.bank_env['dirversion']+'/'+self.bank_env['localrelease']
    env = ''
    if self.bank_env:
      for key, value in self.bank_env.items():
        env += ' -e "{0}={1}"'.format(key, value)
    #         docker run with data.dir env as shared volume
    #         forwarded env variables
    cmd = '''uid={uid}
gid={gid}
{sudo} docker pull {container_id}
{sudo} docker run --rm -w {bank_dir}  -v {data_dir}:{data_dir} {env} {container_id} \
bash -c "groupadd --gid {gid} {group_biomaj} && useradd --uid {uid} --gid {gid} {user_biomaj}; \
{exe} {args}; \
chown -R {uid}:{gid} {bank_dir}"'''.format(uid = os.getuid(),
                                          gid = os.getgid(),
                                          data_dir = self.bank_env['datadir'],
                                          env = env,
                                          container_id = self.docker,
                                          group_biomaj = 'biomaj',
                                          user_biomaj = 'biomaj',
                                          exe = self.exe,
                                          args = ' '.join(self.args),
                                          bank_dir=release_dir,
                                          sudo=use_sudo
                                          )

    (handler, tmpfile) = tempfile.mkstemp('biomaj')
    os.write(handler,cmd)
    os.close(handler)
    os.chmod(tmpfile, 0o755)
    args = [ tmpfile ]
    logging.debug('PROCESS:EXEC:Docker:'+str(self.args))
    logging.debug('PROCESS:EXEC:Docker:Tmpfile:'+tmpfile)
    err= False
    if not simulate:
      logging.info('PROCESS:RUN:Docker:'+self.docker+':'+self.name)
      with open(self.output_file,'w') as fout:
        with open(self.error_file,'w') as ferr:
          if self.expand:
            args = " ".join(args)
            proc = subprocess.Popen(args, stdout=fout, stderr=ferr, env=self.bank_env, shell=True)
          else:
            proc = subprocess.Popen(args, stdout=fout, stderr=ferr, env=self.bank_env, shell=False)
          proc.wait()
          if proc.returncode == 0:
            err = True
          else:
            logging.error('PROCESS:ERROR:'+self.name)
          fout.flush()
          ferr.flush()
    else:
      err = True
    logging.info('PROCESS:EXEC:' + self.name + ':' + str(err))
    os.remove(tmpfile)
    return err


class DrmaaProcess(Process):
  def __init__(self, name, exe, args, desc=None, proc_type=None, native=None, expand=True, bank_env=None, log_dir=None):
    Process.__init__(self, name, exe, args, desc, proc_type, expand, bank_env, log_dir)
    self.native = native


  def run(self, simulate=False):
    '''
    Execute process

    :param simulate: does not execute process
    :type simulate: bool
    :return: exit code of process
    '''
    args = [ self.exe ] + self.args
    logging.debug('PROCESS:EXEC:'+str(self.args))
    err= False
    if not simulate:
      logging.info('Run process '+self.name)
      # Execute on DRMAA
      try:
        import drmaa
        with drmaa.Session() as s:
          jt = s.createJobTemplate()
          jt.remoteCommand = self.exe
          jt.args = self.args
          jt.joinFiles=False
          jt.workingDirectory = os.path.dirname(os.path.realpath(self.output_file))
          jt.jobEnvironment = self.bank_env
          if self.native:
            jt.nativeSpecification = " "+self.native+" "
          jt.outputPath = self.output_file
          jt.errorPath = self.error_file
          jobid = s.runJob(jt)
          retval = s.wait(jobid, drmaa.Session.TIMEOUT_WAIT_FOREVER)
          if  retval.hasExited > 0:
            err = True
          else:
            logging.error('PROCESS:ERROR:'+self.name)
          s.deleteJobTemplate(jt)

      except Exception as e:
        logging.error('Drmaa process error: '+str(e))
        return False
    else:
      err = True
    logging.info('PROCESS:EXEC:' + self.name + ':' + str(err))

    return err



from builtins import str
from builtins import range
from builtins import object
import logging
import datetime
import os
import shutil
import tempfile
import re
import traceback
import json
import time

from biomaj.utils import Utils
from biomaj.download.ftp import FTPDownload
from biomaj.download.http import HTTPDownload
from biomaj.download.direct import MultiDownload, DirectFTPDownload, DirectHttpDownload
from biomaj.download.localcopy import LocalDownload
from biomaj.download.downloadthreads import DownloadThread

from biomaj.mongo_connector import MongoConnector
from biomaj.options import Options

from biomaj.process.processfactory import RemoveProcessFactory,PreProcessFactory,PostProcessFactory

class Workflow(object):
  '''
  Bank update workflow
  '''

  FLOW_INIT = 'init'
  FLOW_CHECK = 'check'
  FLOW_DEPENDS = 'depends'
  FLOW_PREPROCESS = 'preprocess'
  FLOW_RELEASE = 'release'
  FLOW_DOWNLOAD = 'download'
  FLOW_POSTPROCESS = 'postprocess'
  FLOW_REMOVEPROCESS = 'removeprocess'
  FLOW_PUBLISH = 'publish'
  FLOW_OVER = 'over'

  FLOW = [
    { 'name': 'init', 'steps': []},
    { 'name': 'check', 'steps': []},
    { 'name': 'over', 'steps': []}
  ]

  def __init__(self, bank, session=None):
    '''
    Instantiate a new workflow

    :param bank: bank on which to apply the workflow
    :type bank: :class:`biomaj.bank.Bank`
    '''
    self.bank = bank
    if session is None:
      self.session = bank.session
    else:
      self.session = session
      self.bank.session = session
    self.options = bank.options
    self.name = bank.name
    # Skip all remaining tasks, no need to update
    self.skip_all = False

    self.session._session['update'] = False
    self.session._session['remove'] = False

  def get_handler(self, protocol, server, remote_dir, list_file=[]):
    '''
    Get a protocol download handler
    '''
    downloader = None
    if protocol == 'ftp' or protocol == 'sftp':
      downloader = FTPDownload(protocol, server, remote_dir)
    if protocol == 'http':
      downloader = HTTPDownload(protocol, server, remote_dir, self.bank.config)
    if protocol == 'local':
      downloader = LocalDownload(remote_dir)
    if protocol == 'directftp':
      downloader = DirectFTPDownload('ftp', server, remote_dir, list_file)
    if protocol == 'directhttp':
      downloader = DirectHttpDownload('http', server, remote_dir, list_file)
    if downloader is not None:
      downloader.bank = self.bank.name
    return downloader


  def get_flow(self, task):
    for flow in Workflow.FLOW:
      if flow['name'] == task:
        return flow

  def start(self):
    '''
    Start the workflow
    '''
    logging.info('Workflow:Start')
    #print str(self.session._session['status'])
    for flow in self.session.flow:
      if self.skip_all:
        logging.info('Workflow:Skip:'+flow['name'])
        self.session._session['status'][flow['name']] = None
        self.session._session['status'][Workflow.FLOW_OVER] = True
        continue

      if self.options.get_option(Options.STOP_BEFORE) == flow['name']:
        self.wf_over()
        break
      # Always run INIT
      if flow['name'] != Workflow.FLOW_INIT and self.session.get_status(flow['name']):
        logging.info('Workflow:Skip:'+flow['name'])
      if flow['name'] == Workflow.FLOW_INIT or not self.session.get_status(flow['name']):
        logging.info('Workflow:Start:'+flow['name'])
        try:
          self.session._session['status'][flow['name']] = getattr(self, 'wf_'+flow['name'])()
        except Exception as e:
          self.session._session['status'][flow['name']] = False
          logging.error('Workflow:'+flow['name']+'Exception:'+str(e))
          logging.debug(traceback.format_exc())
          #print str(traceback.format_exc())
        finally:
          self.wf_progress(flow['name'], self.session._session['status'][flow['name']])
        if flow['name'] != Workflow.FLOW_OVER and not self.session.get_status(flow['name']):
            logging.error('Error during task '+flow['name'])
            if flow['name'] != Workflow.FLOW_INIT:
              self.wf_over()
            return False
        # Main task is over, execute sub tasks of main
        if not self.skip_all:
          for step in flow['steps']:
            try:
              res = getattr(self, 'wf_'+step)()
              if not res:
                logging.error('Error during '+flow['name']+' subtask: wf_' + step)
                self.wf_over()
                return False
            except Exception as e:
              logging.error('Workflow:'+flow['name']+' subtask: wf_' + step+ ':Exception:'+str(e))
              logging.debug(traceback.format_exc())
              self.wf_over()
              return False
      if self.options.get_option(Options.STOP_AFTER) == flow['name']:
        self.wf_over()
      #if self.options and 'stop_after' in self.options and self.options['stop_after'] == flow['name']:
        break
    self.wf_progress_end()
    return True

  def wf_progress_init(self):
    '''
    Set up new progress status
    '''
    status = {}
    status['log_file'] = {'status': self.session.config.log_file, 'progress': 0}
    status['session'] = self.session._session['id']
    for flow in self.session.flow:
      if flow['name'] == 'download':
        status[flow['name']] = {'status': None, 'progress': 0, 'total': 0}
      elif flow['name'].endswith('process'):
        status[flow['name']] = {'status': None, 'progress': {}}
      elif flow['name'] == 'release':
        status[flow['name']] = {'status': None, 'progress': ''}
      else:
        status[flow['name']] = {'status': None, 'progress': 0}
    MongoConnector.banks.update({'name': self.name},{'$set': {'status': status}})

  def wf_progress_end(self):
    '''
    Reset progress status when workflow is over
    '''
    #MongoConnector.banks.update({'name': self.name},{'$set': {'status': None}})

  def wf_progress(self, task, status):
    '''
    Update bank status
    '''
    subtask = 'status.'+task+'.status'
    MongoConnector.banks.update({'name': self.name},{'$set': {subtask: status}})

  def wf_init(self):
    '''
    Initialize workflow
    '''
    logging.info('Workflow:wf_init')
    data_dir = self.session.config.get('data.dir')
    lock_dir = self.session.config.get('lock.dir', default=data_dir)
    if not os.path.exists(lock_dir):
      os.mkdir(lock_dir)
    lock_file = os.path.join(lock_dir,self.name+'.lock')
    maintenance_lock_file = os.path.join(lock_dir,'biomaj.lock')
    if os.path.exists(maintenance_lock_file):
      logging.error('Biomaj is in maintenance')
      return False
    if os.path.exists(lock_file):
      logging.error('Bank '+self.name+' is locked, a process may be in progress, else remove the lock file '+lock_file)
      #print 'Bank '+self.name+' is locked, a process may be in progress, else remove the lock file'
      return False
    f = open(lock_file, 'w')
    f.write('1')
    f.close()
    self.wf_progress_init()
    return True

  def wf_over(self):
    '''
    Workflow is over
    '''
    logging.info('Workflow:wf_over')
    data_dir = self.session.config.get('data.dir')
    lock_dir = self.session.config.get('lock.dir', default=data_dir)
    lock_file = os.path.join(lock_dir,self.name+'.lock')
    os.remove(lock_file)
    return True

class RemoveWorkflow(Workflow):
  '''
  Workflow to remove a bank instance
  '''

  FLOW = [
    { 'name': 'init', 'steps': []},
    { 'name': 'removeprocess', 'steps': []},
    { 'name': 'remove_release', 'steps': []},
    { 'name': 'over', 'steps': []}
  ]

  def __init__(self, bank, session):
    '''
    Instantiate a new workflow

    :param bank: bank on which to apply the workflow
    :type bank: Bank
    :param session: session to remove
    :type session: :class:`biomaj.session.Session`
    '''
    Workflow.__init__(self, bank, session)
    logging.debug('New workflow')
    self.session._session['remove'] = True


  def wf_remove_release(self):
    logging.info('Workflow:wf_remove_release')
    if not self.session.get('update_session_id'):
      logging.error('Bug: update_session_id not set in session')
      return False

    if os.path.exists(self.session.get_full_release_directory()):
      shutil.rmtree(self.session.get_full_release_directory())
    return self.bank.remove_session(self.session.get('update_session_id'))

  def wf_removeprocess(self):
    logging.info('Workflow:wf_removepreprocess')
    metas = self.session._session['process']['removeprocess']
    pfactory = RemoveProcessFactory(self.bank, metas)
    res = pfactory.run()
    self.session._session['process']['removeprocess'] = pfactory.meta_status
    return res


class UpdateWorkflow(Workflow):
  '''
  Workflow for a bank update
  '''

  FLOW = [
    { 'name': 'init', 'steps': []},
    { 'name': 'check', 'steps': []},
    { 'name': 'depends', 'steps': []},
    { 'name': 'preprocess', 'steps': []},
    { 'name': 'release', 'steps': []},
    { 'name': 'download', 'steps': ['uncompress','copy', 'copydepends']},
    { 'name': 'postprocess', 'steps': ['metadata', 'stats']},
    { 'name': 'publish', 'steps': ['old_biomaj_api', 'clean_offline', 'delete_old', 'clean_old_sessions']},
    { 'name': 'over', 'steps': []}
  ]

  def __init__(self, bank):
    '''
    Instantiate a new workflow

    :param bank: bank on which to apply the workflow
    :type bank: Bank
    '''
    Workflow.__init__(self, bank)
    logging.debug('New workflow')
    self.session._session['update'] = True

  def wf_init(self):
    err = super(UpdateWorkflow, self).wf_init()
    if not err:
      return False
    if self.options.get_option(Options.FROMSCRATCH):
      return self.wf_clean_offline()

    return True

  def wf_check(self):
    '''
    Basic checks
    '''
    logging.info('Workflow:wf_check')
    return True

  def wf_depends(self):
    '''
    Checks bank dependencies with other banks. If bank has dependencies, execute update on other banks first
    '''
    logging.info('Workflow:wf_depends')
    # Always rescan depends, there might be a new release
    self.session.set('depends', {})
    res = self.bank.update_dependencies()
    logging.info('Workflow:wf_depends:'+str(res))
    return res

  def wf_copydepends(self):
    '''
    Copy files from dependent banks if needed
    '''
    logging.info('Workflow:wf_copydepends')
    deps = self.bank.get_dependencies()
    for dep in deps:
      if self.bank.config.get(dep+'.files.move'):
        logging.info('Worflow:wf_depends:Files:Move:'+self.bank.config.get(dep+'.files.move'))
        bdir = None
        for bdep in self.bank.depends:
          if bdep.name == dep:
            bdir = bdep.session.get_full_release_directory()
            break
        if bdir is None:
          logging.error('Could not find a session update for bank '+dep)
          return False
        b = self.bank.get_bank(dep, no_log=True)
        locald = LocalDownload(bdir)
        (file_list, dir_list) = locald.list()
        locald.match(self.bank.config.get(dep+'.files.move').split(), file_list, dir_list)
        bankdepdir = self.bank.session.get_full_release_directory()+"/"+dep
        if not os.path.exists(bankdepdir):
          os.mkdir(bankdepdir)
        downloadedfiles = locald.download(bankdepdir)
        locald.close()
        if not downloadedfiles:
          logging.info('Workflow:wf_copydepends:no files to copy')
          return False
    return True

  def wf_preprocess(self):
    '''
    Execute pre-processes
    '''
    logging.info('Workflow:wf_preprocess')
    metas = self.session._session['process']['preprocess']
    pfactory = PreProcessFactory(self.bank, metas)
    res = pfactory.run()
    self.session._session['process']['preprocess'] = pfactory.meta_status
    return res

  def wf_release(self):
    '''
    Find current release on remote
    '''
    logging.info('Workflow:wf_release')
    cf = self.session.config
    if cf.get('ref.release') and self.bank.depends:
        # Bank is a computed bank and we ask to set release to the same
        # than an other dependant bank
        depbank = self.bank.get_bank(cf.get('ref.release'))
        if depbank and depbank.bank['production']:
          release = depbank.bank['production'][len(depbank.bank['production'])-1]['release']
          self.session.set('release', release)
          self.session.set('remoterelease', release)
          MongoConnector.banks.update({'name': self.bank.name},{'$set': {'status.release.progress': str(release)}})
          return True
        else:
          logging.error('Dependant bank '+str(depbank)+' has no production directory to use')
          return False

    self.session.previous_release = self.session.get('previous_release')
    logging.info('Workflow:wf_release:previous_session:'+str(self.session.previous_release))
    if self.session.get('release'):
      # Release already set from a previous run
      logging.info('Workflow:wf_release:session:'+str(self.session.get('release')))
      return True
    if self.session.config.get('release.file') == '' or self.session.config.get('release.file') is None:
      logging.debug('Workflow:wf_release:norelease')
      self.session.set('release',None)
      return True
    else:
      protocol = cf.get('protocol')
      if cf.get('release.protocol') is not None:
        protocol = cf.get('release.protocol')
      server = cf.get('server')
      if cf.get('release.server') is not None:
        server = cf.get('release.server')
      remote_dir = cf.get('remote.dir')
      if cf.get('release.remote.dir') is not None:
        remote_dir = cf.get('release.remote.dir')
      release_downloader = self.get_handler(protocol,server,remote_dir)
      if cf.get('server.credentials') is not None:
        release_downloader.set_credentials(cf.get('server.credentials'))

      if release_downloader is None:
        logging.error('Protocol '+protocol+' not supported')
        return False

      (file_list, dir_list) = release_downloader.list()

      release_downloader.match([cf.get('release.file')], file_list, dir_list)
      if len(release_downloader.files_to_download) == 0:
        logging.error('release.file defined but does not match any file')
        return False
      if len(release_downloader.files_to_download) > 1:
        logging.error('release.file defined but matches multiple files')
        return False
      if cf.get('release.regexp') is None or not cf.get('release.regexp'):
        # Try to get from regexp in file name
        rel = re.search(cf.get('release.file'), release_downloader.files_to_download[0]['name'])
        if rel is None:
          logging.error('release.file defined but does not match any file')
          return False
        release = rel.group(1)
      else:
        # Download and extract
        tmp_dir = tempfile.mkdtemp('biomaj')
        release_downloader.mkdir_lock = DownloadThread.MKDIR_LOCK
        rel_files = release_downloader.download(tmp_dir)
        rel_file = open(tmp_dir + '/' + rel_files[0]['name'])
        rel_content = rel_file.read()
        rel_file.close()
        shutil.rmtree(tmp_dir)
        rel = re.search(cf.get('release.regexp'), rel_content)
        if rel is None:
          logging.error('release.regexp defined but does not match any file content')
          return False
        # If regexp contains matching group, else take whole match
        if len(rel.groups()) > 0:
          release = rel.group(1)
        else:
          release = rel.group(0)

      release_downloader.close()
      if release_downloader.error:
        logging.error('An error occured during download')
        return False

    self.session.set('release', release)
    self.session.set('remoterelease', release)

    MongoConnector.banks.update({'name': self.bank.name},{'$set': {'status.release.progress': str(release)}})

    # We restart from scratch, a directory with this release already exists
    # Check directory existence if from scratch to change local release
    if self.options.get_option(Options.FROMSCRATCH):
      index = 0
      # Release directory exits, set index to 1
      if os.path.exists(self.session.get_full_release_directory()):
        index = 1
      for x in range(1, 100):
        if os.path.exists(self.session.get_full_release_directory()+'__'+str(x)):
          index = x + 1
      if index > 0:
        self.session.set('release', release+'__'+str(index))
        release = release+'__'+str(index)

    #if self.options.get_option(Options.FROMSCRATCH) and os.path.exists(self.session.get_full_release_directory()):
    #  index = 1
    #  while os.path.exists(self.session.get_full_release_directory()+'_'+str(index)):
    #    index += 1
    #  self.session.set('release', release+'_'+str(index))
    self.download_go_ahead = False
    if self.options.get_option(Options.FROM_TASK) == 'download':
      # We want to download again in same release, that's fine, we do not care it is the same release
      self.download_go_ahead = True
    if not self.download_go_ahead and self.session.previous_release == self.session.get('remoterelease'):
      logging.info('Workflow:wf_release:same_as_previous_session')
      return self.no_need_to_update()

    logging.info('Session:RemoteRelease:'+self.session.get('remoterelease'))
    logging.info('Session:Release:'+self.session.get('release'))
    return True


  def no_need_to_update(self):
    '''
    Set status to over and update = False because there is not a need to update bank
    '''
    self.skip_all = True
    self.session._session['status'][Workflow.FLOW_OVER] = True
    self.session._session['update'] = False
    self.session.set('download_files',[])
    self.wf_over()
    return True

  def wf_download(self):
    '''
    Download remote files or use an available local copy from last production directory if possible.
    '''
    logging.info('Workflow:wf_download')
    flow = self.get_flow(Workflow.FLOW_DOWNLOAD)
    downloader = None
    cf = self.session.config
    self.session.previous_release = self.session.get('previous_release')

    if cf.get('protocol') == 'multi':
      '''
      Search for:
      protocol = multi
      remote.file.0.protocol = directftp
      remote.file.0.server = ftp.ncbi.org
      remote.file.0.path = /musmusculus/chr1/chr1.fa

      => http://ftp2.fr.debian.org/debian/README.html?key1=value&key2=value2
      remote.file.1.protocol = directhttp
      remote.file.1.server = ftp2.fr.debian.org
      remote.file.1.path = debian/README.html
      remote.file.1.method =  GET
      remote.file.1.params.keys = key1,key2
      remote.file.1.params.key1 = value1
      remote.file.1.params.key2 = value2

      => http://ftp2.fr.debian.org/debian/README.html
          #POST PARAMS:
            key1=value
            key2=value2
      remote.file.1.protocol = directhttp
      remote.file.1.server = ftp2.fr.debian.org
      remote.file.1.path = debian/README.html
      remote.file.1.method =  POST
      remote.file.1.params.keys = key1,key2
      remote.file.1.params.key1 = value1
      remote.file.1.params.key2 = value2

      ......
      '''
      downloader = MultiDownload()
      downloaders = []
      # Creates multiple downloaders
      i = 0
      rfile = cf.get('remote.file.'+str(i)+'.path')
      while(rfile is not None):
        if cf.get('remote.file.'+str(i)+'.protocol') is not None:
          protocol = cf.get('remote.file.'+str(i)+'.protocol')
        else:
          protocol = cf.get('protocol')
        if cf.get('remote.file.'+str(i)+'.server') is not None:
          server = cf.get('remote.file.'+str(i)+'.server')
        else:
          server = cf.get('server')
        subdownloader = self.get_handler(protocol,server,'', [cf.get('remote.file.'+str(i)+'.path')])
        if cf.get('remote.file.'+str(i)+'.credentials') is not None:
          credentials = cf.get('remote.file.'+str(i)+'.credentials')
        else:
          credentials = cf.get('server.credentials')
        if credentials is not None:
          subdownloader.set_credentials(credentials)
        if protocol == 'directhttp':
          subdownloader.method = cf.get('remote.file.'+str(i)+'.method')
          if subdownloader.method is None:
            subdownloader.method = 'GET'
          if cf.get('remote.file.'+str(i)+'.name'):
            subdownloader.save_as = cf.get('remote.file.'+str(i)+'.name')
          else:
            subdownloader.save_as = cf.get('remote.file.'+str(i)+'.path')
          if cf.get('remote.file.'+str(i)+'.method'):
            subdownloader.method = cf.get('remote.file.'+str(i)+'.method').strip().upper()
          subdownloader.params = {}
          keys = cf.get('remote.file.'+str(i)+'.params.keys')
          if keys is not None:
            keys = keys.split(',')
            for key in keys:
              param = cf.get('remote.file.'+str(i)+'.params.'+key.strip())
              subdownloader.param[key.strip()] = param.strip()
        downloaders.append(subdownloader)
        i += 1
        rfile = cf.get('remote.file.'+str(i)+'.path')
      downloader.add_downloaders(downloaders)

    else:
      '''
      Simple case, one downloader with regexp
      '''
      protocol = cf.get('protocol')
      if protocol == 'directhttp' or protocol == 'directftp':
        downloader = self.get_handler(cf.get('protocol'),cf.get('server'),'/', [cf.get('remote.dir')[:-1]])
        downloader.method = cf.get('url.method')
        if downloader.method is None:
          downloader.method = 'GET'
        downloader.save_as = cf.get('target.name')
        keys = cf.get('url.params')
        if keys is not None:
          keys = keys.split(',')
          for key in keys:
            param = cf.get(key.strip()+'.value')
            downloader.param[key.strip()] = param.strip()
      else:
        downloader = self.get_handler(cf.get('protocol'),cf.get('server'),cf.get('remote.dir'))

    if downloader is None:
      logging.error('Protocol '+cf.get('protocol')+' not supported')
      return False

    (file_list, dir_list) = downloader.list()

    downloader.match(cf.get('remote.files').split(), file_list, dir_list)
    for f in downloader.files_to_download:
        f['save_as'] = f['name']
        for p in cf.get('remote.files').split():
           res = re.match('/'+p, f['name'])
           if res is not None and res.groups() is not None and len(res.groups())>=1:
             f['save_as'] = '/'.join(res.groups())
             break


    self.session.set('download_files',downloader.files_to_download)
    if self.session.get('release') is None:
        # Not defined, or could not get it ealier
        # Set release to most recent file to download
        release_dict = Utils.get_more_recent_file(downloader.files_to_download)
        if release_dict is None:
            today = datetime.datetime.now()
            release_dict = { 'year': today.year, 'month': today.month, 'day': today.day }

        release = str(release_dict['year']) + '-' + str(release_dict['month']) + '-' + str(release_dict['day'])
        self.session.set('release', release)
        self.session.set('remoterelease', release)
        # We restart from scratch, check if directory with this release already exists
        if self.options.get_option(Options.FROMSCRATCH):
          index = 0
          # Release directory exits, set index to 1
          if os.path.exists(self.session.get_full_release_directory()):
            index = 1
          for x in range(1, 100):
            if os.path.exists(self.session.get_full_release_directory()+'__'+str(x)):
              index = x + 1

          #while os.path.exists(self.session.get_full_release_directory()+'__'+str(index)):
          #  index += 1
          # If we found a directory for this release:   XX or XX__Y
          if index > 0:
            self.session.set('release', release+'__'+str(index))
            release = release+'__'+str(index)
        logging.info('Workflow:wf_download:release:remoterelease:'+self.session.get('remoterelease'))
        logging.info('Workflow:wf_download:release:release:'+release)
        MongoConnector.banks.update({'name': self.bank.name},{'$set': {'status.release.progress': str(release)}})
        self.download_go_ahead = False
        if self.options.get_option(Options.FROM_TASK) == 'download':
          # We want to download again in same release, that's fine, we do not care it is the same release
          self.download_go_ahead = True
        if not self.download_go_ahead and self.session.previous_release == self.session.get('remoterelease'):
          logging.info('Workflow:wf_release:same_as_previous_session')
          return self.no_need_to_update()

    self.banks = MongoConnector.banks
    self.bank.bank = self.banks.find_one({'name': self.name})

    nb_prod_dir = len(self.bank.bank['production'])
    offline_dir = self.session.get_offline_directory()

    copied_files = []

    # Check if already in offlinedir
    keep_files = []
    if os.path.exists(offline_dir):
      for file_to_download in downloader.files_to_download:
        # If file is in offline dir and has same date and size, do not download again
        if os.path.exists(offline_dir + '/' + file_to_download['name']):
          try:
            file_stat = os.stat(offline_dir + '/' + file_to_download['name'])
            f_stat = datetime.datetime.fromtimestamp(os.path.getmtime(offline_dir + '/' + file_to_download['name']))
            year = str(f_stat.year)
            month = str(f_stat.month)
            day = str(f_stat.day)
            if str(file_stat.st_size) != str(file_to_download['size']) or \
               str(year) != str(file_to_download['year']) or \
               str(month) != str(file_to_download['month']) or \
               str(day) != str(file_to_download['day']):
              logging.debug('Workflow:wf_download:different_from_offline:'+file_to_download['name'])
              keep_files.append(file_to_download)
            else:
              logging.debug('Workflow:wf_download:offline:'+file_to_download['name'])
          except Exception as e:
            # Could not get stats on file
            os.remove(offline_dir + '/' + file_to_download['name'])
            keep_files.append(file_to_download)
        else:
          keep_files.append(file_to_download)
      downloader.files_to_download = keep_files

    self.download_go_ahead = False
    if self.options.get_option(Options.FROM_TASK) == 'download':
      # We want to download again in same release, that's fine, we do not care it is the same release
      self.download_go_ahead = True

    if not self.options.get_option(Options.FROMSCRATCH) and not self.download_go_ahead and nb_prod_dir > 0:
      #for prod in self.bank.bank['production']:
      #  if self.session.get('release') == prod['release']:
      #    logging.info('Workflow:wf_release:same_as_previous_production_dir')
      #    return self.no_need_to_update()


      # Get last production
      last_production = self.bank.bank['production'][nb_prod_dir-1]
      # Get session corresponding to production directory
      last_production_session = self.banks.find_one({'name': self.name, 'sessions.id': last_production['session']},{ 'sessions.$': 1})
      last_production_dir = os.path.join(last_production['data_dir'],cf.get('dir.version'),last_production['release'])
      # Checks if some files can be copied instead of downloaded
      downloader.download_or_copy(last_production_session['sessions'][0]['files'],last_production_dir)
      if len(downloader.files_to_download) == 0:
        return self.no_need_to_update()

      #release_dir = os.path.join(self.session.config.get('data.dir'),
      #              self.session.config.get('dir.version'),
      #              self.session.get_release_directory())
      logging.debug('Workflow:wf_download:Copy files from '+last_production_dir)
      copied_files = downloader.files_to_copy
      Utils.copy_files(downloader.files_to_copy, offline_dir)


    downloader.close()

    DownloadThread.NB_THREAD = int(self.session.config.get('files.num.threads'))

    if cf.get('protocol') == 'multi':
      thlist = DownloadThread.get_threads_multi(downloader.downloaders, offline_dir)
    else:
      thlist = DownloadThread.get_threads(downloader, offline_dir)

    running_th = []
    for th in thlist:
      running_th.append(th)
      th.start()

    while len(running_th) > 0:
      try:
          # Join all threads using a timeout so it doesn't block
          # Filter out threads which have been joined or are None
          running_th = [t.join(1000) for t in running_th if t is not None and t.isAlive()]
          logging.debug("Workflow:wf_download:Download:Threads:"+str(running_th))
      except KeyboardInterrupt:
          logging.warn("Ctrl-c received! Sending kill to threads...")
          logging.warn("Running tasks will continue and process will stop.")
          for t in running_th:
              t.downloader.kill_received = True
    logging.info("Workflow:wf_download:Download:Threads:Over")
    #for th in thlist:
    #  th.join()
    is_error = False
    for th in thlist:
      if th.error:
        is_error = True
        downloader.error = True
        break
    self.downloaded_files = downloader.files_to_download + copied_files
    #self.downloaded_files = downloader.download(offline_dir) + copied_files

    #downloader.close()

    if downloader.error:
      logging.error('An error occured during download')
      return False

    return True

  def wf_uncompress(self):
    '''
    Uncompress files if archives and no.extract = false
    '''
    logging.info('Workflow:wf_uncompress')
    no_extract = self.session.config.get('no.extract')
    if no_extract is None or no_extract == 'false':
      for file in self.downloaded_files:
        if 'save_as' not in file:
            file['save_as'] = file['name']
        Utils.uncompress(self.session.get_offline_directory() + '/' + file['save_as'])
    return True

  def wf_copy(self):
    '''
    Copy files from offline directory to release directory
    '''
    logging.info('Workflow:wf_copy')
    from_dir = os.path.join(self.session.config.get('data.dir'),
                  self.session.config.get('offline.dir.name'))
    regexp = self.session.config.get('local.files').split()
    to_dir = os.path.join(self.session.config.get('data.dir'),
                  self.session.config.get('dir.version'),
                  self.session.get_release_directory(), 'flat')

    local_files = Utils.copy_files_with_regexp(from_dir,to_dir,regexp, True)
    self.session._session['files'] = local_files
    if len(self.session._session['files']) == 0:
      logging.error('Workflow:wf_copy:No file match in offline dir')
      return False
    return True

  def wf_metadata(self):
    '''
    Update metadata with info gathered from processes
    '''
    logging.info('Workflow:wf_metadata')
    self.bank.session.set('formats', {})
    per_process_meta_data = self.session.get('per_process_metadata')
    for proc in list(per_process_meta_data.keys()):
      for meta_data in list(per_process_meta_data[proc].keys()):
        session_formats = self.bank.session.get('formats')
        if meta_data not in session_formats:
          #session_formats[meta_data] = [meta_thread.meta_data[meta_data]]
          session_formats[meta_data] = per_process_meta_data[proc][meta_data]
        else:
          #session_formats[meta_data].append(meta_thread.meta_data[meta_data])
          session_formats[meta_data] += per_process_meta_data[proc][meta_data]
    return True

  def wf_stats(self):
    '''
    Get some stats from current release data dir
    '''
    logging.info('Workflow:wf_stats')
    do_stats = self.bank.config.get('data.stats')
    if do_stats is None or do_stats == '0':
      self.session.set('fullsize', 0)
      return True
    prod_dir = self.session.get_full_release_directory()
    dir_size = Utils.get_folder_size(prod_dir)
    self.session.set('fullsize', dir_size)
    return True

  def wf_postprocess(self):
    '''
    Execute post processes
    '''

    # Creates a temporary symlink future_release to keep compatibility if process
    # tries to access dir with this name
    future_link = os.path.join(self.bank.config.get('data.dir'),
                                self.bank.config.get('dir.version'),
                                'future_release')
    prod_dir = self.session.get_full_release_directory()
    to_dir = os.path.join(self.bank.config.get('data.dir'),
                  self.bank.config.get('dir.version'))

    if os.path.lexists(future_link):
      os.remove(future_link)
    os.chdir(to_dir)
    os.symlink(self.session.get_release_directory(),'future_release')

    logging.info('Workflow:wf_postprocess')
    blocks = self.session._session['process']['postprocess']
    pfactory = PostProcessFactory(self.bank, blocks)
    res = pfactory.run()
    self.session._session['process']['postprocess'] = pfactory.blocks

    # In any way, delete symlink
    if os.path.lexists(future_link):
      os.remove(future_link)

    return res

  def wf_publish(self):
    '''
    Add *current* symlink to this release
    '''
    if self.bank.config.get_bool('auto_publish', default=False):
      logging.info('Workflow:wf_publish')
      self.bank.publish()
      return True

    if not self.options.get_option(Options.PUBLISH):
      logging.info('Workflow:wf_publish:no')
      return True
    logging.info('Workflow:wf_publish')
    self.bank.publish()
    return True

  def wf_old_biomaj_api(self):
    '''
    Generates a listing.format file containing the list of files in directories declared in formats
    '''
    release_dir = self.session.get_full_release_directory()
    for release_format in self.bank.session.get('formats'):
      format_file = os.path.join(release_dir,'listingv1.'+release_format.replace('/','_'))
      section = self.list_section(release_dir, release_format, release_format)
      logging.debug("Worfklow:OldAPI:WriteListing: "+format_file)
      fd = os.open( format_file, os.O_RDWR|os.O_CREAT )
      os.write(fd, json.dumps(section).encode('utf-8'))
      os.close( fd )
    return True

  def list_section(self, base_dir, release_format, base_format):
      '''
      Get section files and sub-section from base_dir for directory release_format

      :param base_dir: root directory
      :type base_dir: str
      :param base_dir: sub directory to scan
      :type base_dir: str
      :param base_format: first directroy indicating format
      :type base_format: str
      :return: dict section details
      '''
      section = { "name": release_format, "sections": [], "files": []}
      format_dir = os.path.join(base_dir,release_format)
      if not os.path.exists(format_dir):
        logging.info("Worfklow:OldAPI:Format directory "+release_format+" does not exists, skipping")
        return section
      format_dir_list = os.listdir(format_dir)
      for format_dir_file in format_dir_list:
        if os.path.isfile(os.path.join(format_dir, format_dir_file)):
          if base_format.lower() == 'blast':
            if format_dir_file.endswith('.nal'):
              fileName, fileExtension = os.path.splitext(format_dir_file)
              section['files'].append(os.path.join(format_dir, fileName))
          else:
            section['files'].append(os.path.join(format_dir, format_dir_file))
        else:
          # This is a sub directory
          new_section = self.list_section(format_dir, format_dir_file, base_format)
          section['sections'].append(new_section)
      return section


  def wf_clean_offline(self):
    '''
    Clean offline directory
    '''
    logging.info('Workflow:wf_clean_offline')
    if os.path.exists(self.session.get_offline_directory()):
      shutil.rmtree(self.session.get_offline_directory())
    return True

  def wf_clean_old_sessions(self):
    '''
    Delete old sessions not related to a production directory or last run
    '''
    logging.info('Workflow:wf_clean_old_sessions')
    self.bank.clean_old_sessions()
    return True

  def wf_delete_old(self):
    '''
    Delete old production dirs
    '''
    logging.info('Workflow:wf_delete_old')
    if self.options.get_option(Options.FROM_TASK) is not None:
      # This is a run on an already present release, skip delete
      logging.info('Workflow:wf_delete_old:Skip')
      return True
    keep = int(self.session.config.get('keep.old.version'))
    # Current production dir is not yet in list
    nb_prod = len(self.bank.bank['production'])
    # save session during delete workflow
    keep_session = self.bank.session

    if nb_prod > keep:
      for prod in self.bank.bank['production']:
        if 'freeze' in prod and prod['freeze']:
          continue
        if self.bank.bank['current'] == prod['session']:
          continue
        if nb_prod - keep > 0:
          nb_prod -= 1
          session = self.bank.get_new_session(RemoveWorkflow.FLOW)
          # Delete init and over because we are already in a run
          i_init = -1
          i_over = -1
          for i in range(0,len(session.flow)):
            if session.flow[i]['name'] == 'init':
              i_init = i
          if i_init >= 0:
            del session.flow[i_init]
          for i in range(0,len(session.flow)):
            if session.flow[i]['name'] == 'over':
              i_over = i
          if i_over >= 0:
            del session.flow[i_over]

          session.set('action', 'remove')
          session.set('release', prod['release'])
          session.set('remoterelease', prod['remoterelease'])
          session.set('update_session_id', prod['session'])
          logging.info('Workflow:wf_delete_old:Delete:'+prod['release'])
          res = self.bank.start_remove(session)
          if not res:
            logging.error('Workflow:wf_delete_old:ErrorDelete:'+prod['release'])
        else:
          break
    # Set session back
    self.bank.session  = keep_session

    return True



from builtins import str
from builtins import object
import smtplib
import email.utils
import sys
if sys.version < '3':
    from email.MIMEText import MIMEText
else:
    from email.mime.text import MIMEText

from biomaj.workflow import Workflow
import logging

class Notify:
  '''
  Send notifications
  '''

  @staticmethod
  def notifyBankAction(bank):
    if not bank.config.get('mail.smtp.host') or bank.session is None:
      logging.info('Notify:none')
      return
    logging.info('Notify:'+bank.config.get('mail.admin'))
    mfrom = bank.config.get('mail.from')
    mto = bank.config.get('mail.admin')
    log_file = bank.config.log_file
    msg = MIMEText('')
    if log_file:
      fp = open(log_file, 'rb')
      msg = MIMEText(fp.read())
      fp.close()
    msg['To'] = email.utils.formataddr(('Recipient', mto))
    msg['From'] = email.utils.formataddr(('Author', mfrom))
    #msg['Subject'] = 'BANK['+bank.name+'] - STATUS['+str(bank.session.get_status(Workflow.FLOW_OVER))+'] - UPDATE['+str(bank.session.get('update'))+'] - REMOVE['+str(bank.session.get('remove'))+']'
    msg['Subject'] = 'BANK['+bank.name+'] - STATUS['+str(bank.session.get_status(Workflow.FLOW_OVER))+'] - UPDATE['+str(bank.session.get('update'))+'] - REMOVE['+str(bank.session.get('remove'))+']' + ' - RELEASE['+str(bank.session.get('release'))+']'
    #if bank.session.get('action') == 'update':
    #  msg['Subject'] = 'BANK['+bank.name+'] - STATUS['+str(bank.session.get_status(Workflow.FLOW_OVER))+'] - UPDATE['+str(bank.session.get('update'))+'] - REMOVE['+str(bank.session.get('remove'))+']' + ' - RELEASE['+str(bank.session.get('release'))+']'
    #else:
    #  msg['Subject'] = 'BANK['+bank.name+'] - STATUS['+str(bank.session.get_status(Workflow.FLOW_OVER))+'] - UPDATE['+str(bank.session.get('update'))+'] - REMOVE['+str(bank.session.get('remove'))+']'
    logging.info(msg['subject'])
    server = None
    try:
      server = smtplib.SMTP(bank.config.get('mail.smtp.host'))
      #server.set_debuglevel(1)
      if bank.config.get('mail.tls') is not None and str(bank.config.get('mail.tls')) == 'true':
        server.starttls()
      if bank.config.get('mail.user') is not None and str(bank.config.get('mail.user')) != '':
        server.login(bank.config.get('mail.user'),bank.config.get('mail.password'))
      server.sendmail(mfrom, [mto], msg.as_string())
    except Exception as e:
      logging.error('Could not send email: '+str(e))
    finally:
      if server is not None:
        server.quit()



from builtins import object


class Options(object):
  '''
  Available options
  '''

  def __init__(self, options=None):
      self.options = options

  def get_option(self, option):
    '''
    Gets an option if present, else return None
    '''
    #if self.options is None:
    #  return None

    #if hasattr(self.options, option):
    #  return getattr(self.options, option)

    if hasattr(self, option):
      return getattr(self, option)
    #if option in self.options:
    #  return self.options[option]

    return None

  UPDATE = 'update'
  REMOVE = 'remove'
  PUBLISH = 'publish'
  FROM_TASK = 'from_task'
  PROCESS = 'process'
  STOP_BEFORE = 'stop_before'
  STOP_AFTER = 'stop_after'
  FROMSCRATCH = 'fromscratch'



from builtins import str
from builtins import object
import logging
import copy
from elasticsearch import Elasticsearch

class BmajIndex(object):
    '''
    ElasticSearch indexation and search
    '''


    '''
    ElasticSearch server
    '''
    es = None

    '''
    Index name
    '''
    index = 'biomaj'

    '''
    Do indexing
    '''
    do_index = False

    @staticmethod
    def load(hosts=['localhost'], index='biomaj', do_index=True):
      '''
      Initialize index

      :param hosts: List of elastic search nodes to connect to
      :type hosts: list
      :param do_index: index data or not
      :type do_index: bool
      '''
      if not do_index:
        return
      BmajIndex.index = index
      BmajIndex.do_index = do_index
      if BmajIndex.es is None:
        BmajIndex.es = Elasticsearch(hosts)

        mapping = {
          "mappings": {
                "production": {
                  "date_detection": False
                },
                "releasestats": {
                  "date_detection": False,
                  "_timestamp" : {
                    "enabled" : True,
                    "store" : True
                  }
                }
            }
        }
        try:
          if not BmajIndex.es.indices.exists(index=BmajIndex.index):
            BmajIndex.es.indices.create(index=BmajIndex.index,body=mapping)
        except Exception as e:
          logging.error('ElasticSearch connection error, check server is running and configuration')
          raise e


    @staticmethod
    def delete_all_bank(bank_name):
      '''
      Delete complete index for a bank
      '''
      if not BmajIndex.do_index:
        return
      query = {
        "query" : {
          "term" : { "bank" : bank_name }
          }
        }
      BmajIndex.es.delete_by_query(index=BmajIndex.index, body=query)

    @staticmethod
    def remove(bank_name, release):
      '''
      Remove a production release

      :param bank_name: Name of the bank
      :type bank_name: str
      :param release: production release
      :type release: str
      '''
      if not BmajIndex.do_index:
        return
      try:
        query = {
          "query" : {
            "term" : { "release" : release, "bank": bank_name }
            }
          }
        BmajIndex.es.delete_by_query(index=BmajIndex.index, body=query)
      except Exception as e:
        logging.error('Index:Remove:'+bank_name+'_'+str(release)+':Exception:'+str(e))

    @staticmethod
    def search(query):
      if not BmajIndex.do_index:
        return None
      res = BmajIndex.es.search(index=BmajIndex.index, doc_type='production', body=query)
      return res['hits']['hits']

    @staticmethod
    def searchq(query,size=1000):
      '''
      Lucene syntax search

      :param query: Lucene search string
      :type query: str
      :param size: number of results
      :type size: int
      :return: list of matches
      '''
      if not BmajIndex.do_index:
        return None
      res = BmajIndex.es.search(index=BmajIndex.index, doc_type='production', q=query, size=size)
      return res['hits']['hits']

    @staticmethod
    def add_stat(stat_id, stat):
      '''
      Add some statistics, must contain release and bank properties.
      '''
      if not BmajIndex.do_index:
        return
      if stat['release'] is None or stat['bank'] is None:
        return False
      obj['bank'] = bank_name

      BmajIndex.es.index(index=BmajIndex.index, doc_type='releasestats', id=stat_id, body=stat)
      return True


    @staticmethod
    def add(bank_name, prod, flush=False):
      '''
      Index a production release

      :param bank_name: Name of the bank
      :type bank_name: str
      :param prod: session release object
      :type prod: dict
      :param flush: Force flushing
      :type flush: bool
      '''
      if not BmajIndex.do_index:
        return
      obj = copy.deepcopy(prod)
      if obj['release'] is None:
        return
      obj['bank'] = bank_name
      formats = obj['formats']
      try:
        for fkey,fvalue in formats.items():
          for elt in fvalue:
            elt['format'] = fkey
            elt['bank'] = bank_name
            elt['release'] = obj['release']
            if 'status' in obj:
              elt['status'] = obj['status']
            res = BmajIndex.es.index(index=BmajIndex.index, doc_type='production', body=elt)
        if flush:
          BmajIndex.es.indices.flush(index=BmajIndex.index, force=True)
      except Exception as e:
        logging.error('Index:Add:'+bank_name+'_'+str(obj['release'])+':Exception:'+str(e))



from builtins import str
from builtins import object
import tarfile
import zipfile
import gzip
import bz2
import re
import glob
import os
import logging
import shutil
import datetime
import subprocess

from mimetypes import MimeTypes

class Utils(object):
  '''
  Utility classes
  '''

  mime = None

  @staticmethod
  def get_folder_size(folder):
    '''
    Get directory path full size

    :param folder: directory path
    :type folder: str
    '''
    if not os.path.exists(folder):
      return -1
    folder_size = 0
    for (path, dirs, files) in os.walk(folder):
      for file in files:
        filename = os.path.join(path, file)
        folder_size += os.path.getsize(filename)
    return folder_size


  @staticmethod
  def detect_format(filename):
    '''
    try to detect file format by extension
    '''
    if Utils.mime is None:
      Utils.mime = MimeTypes()
      mimesfile = os.path.join(os.path.dirname(__file__),'mimes-bio.txt')
      Utils.mime.read(mimesfile,True)
    return Utils.mime.guess_type(filename, True)

  @staticmethod
  def get_more_recent_file(files):
    '''
    Return the date of the most recent file in list.

    Each file is a dict like with (at least) parameters: year, month, day
    '''
    release = None
    for file in files:
      if release is None:
        release = { 'year': file['year'], 'month': file['month'], 'day': file['day']}
      else:
        rel_date = datetime.date(int(release['year']),int(release['month']),int(release['day']))
        file_date = datetime.date(int(file['year']),int(file['month']),int(file['day']))
        if file_date > rel_date:
          release['year'] = file['year']
          release['month'] = file['month']
          release['day'] = file['day']
    return release

  @staticmethod
  def month_to_num(date):
    return{
          'Jan' : 1,
          'Feb' : 2,
          'Mar' : 3,
          'Apr' : 4,
          'May' : 5,
          'Jun' : 6,
          'Jul' : 7,
          'Aug' : 8,
          'Sep' : 9,
          'Oct' : 10,
          'Nov' : 11,
          'Dec' : 12
          }[date]

  @staticmethod
  def copy_files(files_to_copy, to_dir, move=False, lock=None):
    '''
    Copy or move files to to_dir, keeping directory structure.

    Copy keeps the original file stats.
    Files should have attributes name and root:
    - root: root directory
    - name: relative path of file in root directory

    /root/file/file1 will be copied in to_dir/file/file1

    :param files_to_copy: list of files to copy
    :type files_to_copy: list
    :param to_dir: destination directory
    :type to_dir: str
    :param move: move instead of copy
    :type move: bool
    :param lock: thread lock object for multi-threads
    :type lock: Lock
    '''
    nb_files = len(files_to_copy)
    cur_files = 1
    for file_to_copy in files_to_copy:
      logging.debug(str(cur_files)+'/'+str(nb_files)+' copy file '+file_to_copy['name'])
      cur_files += 1
      from_file = file_to_copy['root'] + '/' + file_to_copy['name']
      to_file = to_dir + '/' + file_to_copy['name']
      if lock is not None:
        lock.acquire()
        try:
          if not os.path.exists(os.path.dirname(to_file)):
            os.makedirs(os.path.dirname(to_file))
        except Exception as  e:
          logging.error(e)
        finally:
          lock.release()

      else:
        if not os.path.exists(os.path.dirname(to_file)):
          try:
            os.makedirs(os.path.dirname(to_file))
          except Exception as  e:
            logging.error(e)
      if move:
        shutil.move(from_file, to_file)
      else:
        shutil.copyfile(from_file, to_file)
        shutil.copystat(from_file, to_file)

  @staticmethod
  def copy_files_with_regexp(from_dir, to_dir, regexps, move=False, lock=None):
    '''
    Copy or move files from from_dir to to_dir matching regexps.
    Copy keeps the original file stats.

    :param from_dir: origin directory
    :type from_dir: str
    :param to_dir: destination directory
    :type to_dir: str
    :param regexps: list of regular expressions that files in from_dir should match to be copied
    :type regexps: list
    :param move: move instead of copy
    :type move: bool
    :param lock: thread lock object for multi-threads
    :type lock: Lock
    :return: list of copied files with their size
    '''
    #os.chdir(from_dir)
    files_to_copy = []
    for root, dirs, files in os.walk(from_dir, topdown=True):
      for name in files:
        for reg in regexps:
          file_relative_path = os.path.join(root, name).replace(from_dir,'')
          if file_relative_path.startswith('/'):
            file_relative_path = file_relative_path.replace('/', '', 1)
          if re.match(reg, file_relative_path):
            files_to_copy.append({'name': file_relative_path})
            continue

    for file_to_copy in files_to_copy:
      from_file = from_dir +'/' + file_to_copy['name']
      to_file = to_dir + '/' + file_to_copy['name']

      if lock is not None:
        lock.acquire()
        try:
          if not os.path.exists(os.path.dirname(to_file)):
            os.makedirs(os.path.dirname(to_file))
        except Exception as  e:
          logging.error(e)
        finally:
          lock.release()
      else:
        if not os.path.exists(os.path.dirname(to_file)):
          os.makedirs(os.path.dirname(to_file))
      if move:
        shutil.move(from_file, to_file)
      else:
        shutil.copyfile(from_file, to_file)
        shutil.copystat(from_file, to_file)
      file_to_copy['size'] = os.path.getsize(to_file)
      f_stat = datetime.datetime.fromtimestamp(os.path.getmtime(to_file))
      file_to_copy['year'] = str(f_stat.year)
      file_to_copy['month'] = str(f_stat.month)
      file_to_copy['day'] = str(f_stat.day)
      (file_format, encoding) = Utils.detect_format(to_file)
      file_to_copy['format'] = file_format
    return files_to_copy

  @staticmethod
  def uncompress(file, remove=True):
    '''
    Test if file is an archive, and uncompress it
    Remove archive file if specified

    :param file: full path to file to check and uncompress
    :type file: str
    :param remove: remove archive if present
    :type remove: bool
    '''
    is_archive = False
    #if tarfile.is_tarfile(file):
    #  logging.debug('Uncompress:Tar:'+file)
    #  tfile = tarfile.TarFile(file)
    #  tfile.extractall(os.path.basename(file))
    #  tfile.close()
    #  is_archive = True
    if file.endswith('.tar.gz'):
        proc = subprocess.check_call("tar xfz "+file+" --overwrite -C "+os.path.dirname(file), shell=True)
        #proc.wait()
        is_archive = True
    elif file.endswith('.tar'):
        proc = subprocess.check_call("tar xf "+file+" --overwrite -C "+os.path.dirname(file), shell=True)
        #proc.wait()
        is_archive = True
    elif file.endswith('.bz2'):
        proc = subprocess.check_call("tar xjf "+file+" --overwrite -C "+os.path.dirname(file), shell=True)
        #proc.wait()
        is_archive = True
    elif file.endswith('.gz'):
        proc = subprocess.check_call("gunzip -f "+file, shell=True)
        #proc.wait()
        is_archive = True
    elif file.endswith('.zip'):
        proc = subprocess.check_call("unzip -o "+file+" -d "+os.path.dirname(file), shell=True)
        #proc.wait()
        is_archive = True
    #elif zipfile.is_zipfile(file):
    #  logging.debug('Uncompress:Zip:'+file)
    #  zfile = zipfile.ZipFile(file)
    #  zfile.extractall(os.path.basename(file))
    #  zfile.close()
    #  is_archive = True
    #elif file.endswith('.gz'):
    #  logging.debug('Uncompress:Gz:'+file)
    #  f_in = open(file.replace('.gz',''), 'wb')
    #  gz_file = gzip.GzipFile(file)
    #  f_in.writelines(gz_file.readlines())
    #  f_in.close()
    #  gz_file.close()
    #  is_archive = True
    #elif file.endswith('.bz2'):
    #  logging.debug('Uncompress:Bz2:'+file)
    #  f_in = open(file.replace('.bz2',''), 'wb')
    #  bz_file = bz2.BZ2File(file)
    #  f_in.writelines(bz_file.readlines())
    #  f_in.close()
    #  bz_file.close()
    #  is_archive = True

    if is_archive:
      logging.debug('Uncompress:uncompress:'+file)


    if is_archive and remove and os.path.exists(file):
      os.remove(file)
